"""
File Purpose: derivatives (in one dimension at a time).
"""
import numpy as np
import xarray as xr

from ..quantity_loader import QuantityLoader
from ...errors import QuantCalcError, InputError
from ...tools import (
    simple_property,
    xarray_differentiate, xarray_isel, interprets_fractional_indexing,
)

DVARS = 'xyzt'  # one-letter string coordinate name options for differentiation.

class BasicDerivativeLoader(QuantityLoader):
    '''derivatives (in one dimension at a time).
    E.g. 'ddx_n' --> derivative of n with respect to x
    '''
    cls_behavior_attrs.register('deriv_before_slice', default='step')
    _VALID_DERIV_BEFORE_SLICE_MODES = (True, False, 'step')
    deriv_before_slice = simple_property('_deriv_before_slice', default='step',
        validate_from='_VALID_DERIV_BEFORE_SLICE_MODES',
        doc='''bool or 'step'. Whether to apply derivatives before any maindims slicing.
        see help(type(self).slices) for slicing details.

        True --> get derivatives before applying slices.
                E.g. if slices=dict(x=slice(10, 30, 4)), getting ddx_var,
                    would compute var across the entire domain, then ddx, then slice by x[10:30:4].
                This mode is most computationally expensive, but derivatives are trustworthy everywhere.
        'step' --> get derivatives before slicing step, but after slicing start & stop.
                E.g. if slices=dict(x=slice(10, 30, 4)), getting ddx_var,
                    would compute var from x[10] to x[30], then ddx, then slice by x[::4].
                Note: if slicing by non-slice (e.g. x=5, or y=[0, 7, 10]), apply slicing afterwards.
                This mode balances computational cost and trustworthiness;
                    derivatives are trustworthy everywhere except near slices' start & stop.
        False --> get derivatives after applying slices.
                E.g. if slices=dict(x=slice(10, 30, 4)), getting ddx_var,
                    would compute var from x[10:30:4], then ddx.
                This mode is computationally cheapest, but derivatives are least trustworthy.''')

    def _pre_and_post_deriv_slices(self, x):
        '''return slices to apply before & after differentiation. (result is a 2-tuple of dicts.)

        x: str, or list of str.
            only slices relevant to these dimensions will be adjusted.
            (e.g. there's no need to touch 'y' slices during ddx.)

        Result depends primarily on self.deriv_before_slice.
        Note if self.deriv_before_slice=False, this method just returns (self.slices, {}).
        '''
        if isinstance(x, str):
            x = [x]
        mode = self.deriv_before_slice
        if mode == 'step':  # slice steps afterwards, slice other things beforehand.
            pre = self.slices.copy()
            post = {}
            maindims_sizes = None
            for dim in x:
                if dim in pre:
                    ss = pre[dim]
                    if isinstance(ss, slice):
                        if ss.step is not None and ss.step != 1:
                            if (-1 < ss.step < 1) and ss.step != 0:  # i.e., step is fractional
                                # must interpret fractional indexing before splitting into pre & post.
                                # otherwise, e.g., if didn't do it, (0, 200, 0.1), with len=500, implies 0.1 <--> step=50;
                                # but splitting to (0, 200, None), then (None, None, 0.1), implies 0.1 <--> step=20,
                                # causing result shape to depend on deriv_before_slice mode, which is not desireable...
                                if maindims_sizes is None:
                                    maindims_sizes = self.maindims_full_sizes  # sizes without slices
                                ss = interprets_fractional_indexing(ss, maindims_sizes[dim])
                        if ss.step is not None and ss.step != 1:
                            pre[dim] = type(ss)(ss.start, ss.stop, None)
                            post[dim] = type(ss)(None, None, ss.step)
                    else:  # non-slice ss: apply it afterwards.
                        del pre[dim]
                        post[dim] = ss
        elif mode == True:  # slice everything (relevant to x) afterwards.
            pre = self.slices.copy()
            post = {}
            for dim in x:
                if dim in pre:
                    post[dim] = pre.pop(dim)
        elif mode == False:  # slice everything beforehand.
            pre = self.slices
            post = {}
        else:
            assert False, f'coding error; invalid mode ({mode!r}) should be prevented by deriv_before_slice setter.'
        return (pre, post)

    @known_pattern(fr'dd([{DVARS}])_(.+)', deps=[1])  # 'dd{x}_{var}'
    def get_derivative(self, var, *, _match=None):
        '''derivative. dd{x}_{var} --> d{var}/d{x}.
        self(var) must return an object with {x} in its coordinates.
            E.g. for x='y', self(var).coords['y'] are required.
        '''
        x, var = _match.groups()
        pre_slices, post_slices = self._pre_and_post_deriv_slices(x)
        with self.using(slices=pre_slices):
            val0 = self(var)
        result = xarray_differentiate(val0, x)
        if post_slices:
            result = xarray_isel(result, **post_slices)
        return result



"""
File Purpose: calculating timescales, e.g. timescale from plasma oscillations.
(e.g., useful when trying to pick dt for a simulation)
"""
import numpy as np

from .quantity_loader import QuantityLoader
from ..errors import QuantCalcError, InputMissingError
from ..tools import (
    UNSET,
    xarray_min, xarray_minimum_of_datavars, xarray_varmin,
)

class TimescalesLoader(QuantityLoader):
    '''timescales, e.g. timescale from wplasma, or from vthermal / dx.
    Spatial scales come from self('ds')
        which must be defined elsewhere (e.g. in the relevant BasesLoader),
        otherwise trying to get the related timescales will cause a crash.
    '''
    # # # ALL TIMESCALES # # #
    TIMESCALE_VARS = [
        'timescale_wplasma', 'timescale_gyrof', 'timescale_nusn',  # raw timescales
        'timescale_vtherm', 'timescale_EBdrift',  # timescale from speed & ds
    ]

    @known_var(deps=TIMESCALE_VARS)
    def get_timescales(self):
        '''all timescales (from self.TIMESCALE_VARS) as a dataset.
        Consider also: self('timescales_abbrv')

        Useful patterns you might want to consider:
            self('timescales', maindims_means=True)  # timescales based on mean values only
            self('timescales').min('fluid')          # minimum timescales across all fluids
            self('timescales').pc.minimum()          # minimum timescale across all timescales
                equivalent: self('min_timescale')
            self('timescales').min('fluid').pc.minimum()  # minimum timescale at each point.
            self('timescales').pc.varmin()       # name of timescale with minimum value at each point.

            # timescale variable names, sorted from min to max values.
            self('timescales').to_dataarray().pc.sort_along('variable')['variable']
        '''
        timescale_vars = self.TIMESCALE_VARS
        return self(timescale_vars)

    @known_var(deps=TIMESCALE_VARS)
    def get_timescales_abbrv(self):
        '''all timescales (from self.TIMESCALE_VARS) as a dataset, abbreviating names.
        abbreviates 'timescale_var' --> 'var'.
        '''
        result = self('timescales')
        newnames = {v: v[len('timescale_'):] if v.startswith('timescale_') else v for v in result.data_vars}
        return result.rename(newnames)

    @known_var(deps=['timescales'])
    def get_min_timescale(self):
        '''minimum timescale across all timescales.
        Equivalent: self('timescales').pc.minimum()
        '''
        return xarray_minimum_of_datavars(self('timescales'))

    @known_var(deps=['timescales'])
    def get_minf_timescale(self):
        '''minimum timescale across all timescales and fluids.
        Equivalent: self('timescales').min('fluid').pc.minimum()
        '''
        minf = xarray_min(self('timescales'), 'fluid')
        return xarray_minimum_of_datavars(minf)

    @known_var(deps=['timescales_abbrv'])
    def get_min_timescale_type(self):
        '''tells which timescale has the minimum value (across all timescales).
        Equivalent: self('timescales_abbrv').pc.varmin()

        result is a string-valued array, values from self.TIMESCALE_VARS (removing 'timescale_' in names):
            'wplasma', 'gyrof', 'nusn', 'vtherm', 'EBspeed'
        '''
        return xarray_varmin(self('timescales_abbrv'))

    @known_var(deps=['timescales_abbrv'])
    def get_minf_timescale_type(self):
        '''tells which timescale has the minimum value (across all timescales and fluids).
        Equivalent: self('timescales_abbrv').min('fluid').pc.varmin()

        result is a string-valued array, values from self.TIMESCALE_VARS (removing 'timescale_' in names):
            'wplasma', 'gyrof', 'nusn', 'vtherm', 'EBspeed'
        '''
        minf = xarray_min(self('timescales_abbrv'), 'fluid')
        return xarray_varmin(minf)

    # # # HELPERS # # #
    @known_var(deps=['ds'])
    def get_ds_for_timescales(self):
        '''spatial scale used when calculating timescales. vector(ds), e.g. [dx, dy, dz].
        The method here just returns ds. Subclasses might overwrite if they use a different ds for timescales.
        '''
        return self('ds')
    
    @known_var(deps=['ds_for_timescales'])
    def get_dsmin_for_timescales(self):
        '''minimum spatial scale used when calculating timescales.
        min(ds_for_timescales) across components.
        '''
        return xarray_min(self('ds_for_timescales'), 'component', missing_dims='ignore')

    # # # TIMESCALES # # #
    @known_var(deps=['wplasma'])
    def get_timescale_wplasma(self):
        '''timescale from plasma oscillations. 2 * pi / wplasma.  (Hz, not rad/s)
        wplasma = sqrt(n q^2 / (m epsilon0)).
        '''
        return 2 * np.pi / self('wplasma')

    @known_var(deps=['gyrof'])
    def get_timescale_gyrof(self):
        '''timescale for cyclotron motion. 2 * pi / gyrof.  (Hz, not rad/s)
        gyrof = |q| |B| / m.
        '''
        return 2 * np.pi / self('gyrof')

    @known_var(deps=['nusn'])
    def get_timescale_nusn(self):
        '''timescale for collisions with neutrals. 1 / nusn.
        nusn = collision frequency of self.fluid with neutrals.
        '''
        return 1 / self('nusn')

    @known_var(deps=['dsmin_for_timescales', 'vthermal'], aliases=['timescale_vth', 'timescale_vthermal'])
    def get_timescale_vtherm(self):
        '''timescale from thermal velocity. dsmin / vthermal.
        vthermal = sqrt(kB T / m).
        '''
        return self('dsmin_for_timescales') / self('vthermal')

    @known_var(deps=['dsmin_for_timescales', 'E_un0_perpmod_B', 'mod_B'])
    def get_timescale_EBspeed(self):
        '''timescale from speed using E & B fields. dsmin / (|E_un0 cross B| / |B|^2).
        E_un0 (not E) because the derivation assumes neutral frame: u_neutral=0.

        note: to be more precise, use timescale_udrift instead.
        '''
        return self('dsmin_for_timescales') / (self('E_un0_perpmod_B') / self('mod_B'))

    @known_var(deps=['dsmin_for_timescales', 'u_drift'])
    def get_timescale_udrift(self):
        '''timescale from drift speed. dsmin / mod_u_drift.

        note: to be less precise (but computationally cheaper), use timescale_EBspeed instead.
        For electrons when kappae >> 1, timescale_EBspeed ~= timescale_udrift.
        When accounting for directionality, or kappae <~= 1, or non-electrons,
            timescale_EBspeed is always more conservative (i.e. smaller) than timescale_udrift.
        '''
        return self('dsmin_for_timescales') / self('mod_u_drift')

    @known_var(deps=['dsmin_for_timescales', 'timescale_udrift', 'timescale_EBspeed'])
    def get_timescale_EBdrift(self):
        '''timescale from drift speed if possible, else from speed using E & B fields.
        tries to return self('timescale_udrift'), but if that causes a QuantCalcError,
            use self('timescale_EBspeed') instead.
        '''
        try:
            return self('timescale_udrift')
        except QuantCalcError:
            return self('timescale_EBspeed')

    @known_var(deps=['min_timescale'])
    def get_best_subcycle(self):
        '''largest subcycling allowed: best_subcycle = min_timescale / minf_timescale.
        min_timescale = min timescale for self.fluid, across all timescales.
        minf_timescale = min timescale across all fluids and all timescales.
                (when computing minf_timescale here, always use all self.fluids.)
        E.g. fluid=fluids=['e','H+','C+'], min_timescale=[1e-8, 1e-7, 5e-7]
            --> minf_timescale will be 1e-8, so result will be [1, 10, 50].
        '''
        min_timescale = self('min_timescale')
        minf_timescale = self('minf_timescale', fluid=None)
        return min_timescale / minf_timescale

    @known_var(deps=['best_subcycle'])
    def get_best_pow2_subcycle(self):
        '''largest power of 2 subcycling allowed for each fluid in self.fluid.
        result = array of values like 2^N, with largest N such that result < best_subcycle.
        '''
        return 2 ** np.log2(self('best_subcycle')).astype('int')

    @known_var(deps=['best_subcycle'])
    def get_safe_pow2_subcycle(self, *, subcycle_safety=UNSET):
        '''largest "safe" power of 2 subcycling allowed for each fluid in self.fluid.
        result = array of values like 2^N, with largest N such that result < best_subcycle / safety.
        (larger safety produces "safer" results. For safe_subcycle, just use best_subcycle / safety.)
        '''
        if hasattr(self, '_cached_safe_pow2_subcycle'):  # [EFF] if internally getting safe_pow2_subcycle many times,
            return self._cached_safe_pow2_subcycle       # can do "with self.using(_cached_safe_pow2_subcycle=...):"
        if subcycle_safety is UNSET:
            if hasattr(self, 'subcycle_safety'):
                subcycle_safety = self.subcycle_safety
            else:
                raise InputMissingError('subcycle_safety must be set in self or entered as kwarg.')
        if subcycle_safety is None:
            subcycle_safety = 1
        return 2 ** np.log2(self('best_subcycle') / subcycle_safety).astype('int')



"""
File Purpose: trees
"""

from IPython.display import display, HTML

from .display import repr_simple
from .properties import (
    weakref_property_simple, simple_property,
)
from ..defaults import DEFAULTS

''' --------------------- Tree --------------------- '''

class Tree():
    '''node in a Tree. Might have any number of children, might have one parent.
    obj: any object.
        the object stored here, in this node.
    '''
    _assert_type=None  # if not None, assert isinstance(obj, _assert_type) when creating new nodes.

    def __init__(self, obj, parent=None):
        if self._assert_type is not None:
            if not isinstance(obj, self._assert_type):
                raise TypeError(f'obj must be an instance of {self._assert_type}, but got type(obj)={type(obj)}')
        self.obj = obj
        self.children = []
        self.parent = parent   # property; self.parent=parent calls self.set_parent(parent).

    # # # PROPERTIES # # #
    @property
    def parent(self):
        '''parent node of self. None if self is root.
        When set to a value, calls self.set_parent(value),
            which also updates tracking info appropriately, and updates parent's children.
        '''
        return self.parent_ref
    @parent.setter
    def parent(self, value):
        self.set_parent(value)

    parent_ref = weakref_property_simple('_parent_ref',
            doc='''stores parent value, but internally uses weakref to avoid circular references.
            Users should always use self.parent instead.''')

    depth = simple_property('_depth', default=0,
            doc='''number of layers above self. (parent, parent's parents, etc.)
            depth = 0 for the node with parent = None.''')
    height = simple_property('_height', default=0,
            doc='''number of layers below self. (children, children's children, etc.)
            height = 0 for a node with no children.''')
    size = simple_property('_size', default=1,
            doc='''number of nodes in this tree. (here and below)
            size = 1 for a node with no children.''')

    # # # TRACKING INFO WHEN ADDING CHILDREN OR PARENTS # # #
    def set_parent(self, parent, *, _internal=False):
        '''sets self.parent = parent. Also, parent.add_child(self), unless _internal=True.
        Users should use self.parent = parent instead of calling set_parent directly.
        '''
        self._on_setting_parent(parent)
        self.parent_ref = parent
        if (not _internal) and (parent is not None):
            parent.add_child(self)

    def add_child(self, child):
        '''adds this child (a Tree) to self.children. Also, child.set_parent(self).
        returns the added child.
        '''
        self.children.append(child)
        child.set_parent(self,
                _internal=True,  # avoids recursion from add_child inside set_parent.
                )
        self._on_added_child(child)
        return child

    def _on_setting_parent(self, parent):
        '''called immediately before setting self.parent = parent.
        Tracks depth, height, size. This method does not connect any parents or children.
        '''
        self_parent = getattr(self, 'parent', None)  # getattr -> even works before setting self.parent!
        if self_parent is parent:
            return  # don't need to do anything, didn't update parent at all!
        elif self_parent is not None:
            raise NotImplementedError('swapping parents (after setting parent = non-None value)!')
        self.depth = 0 if parent is None else parent.depth + 1
        for child in self.children:
            child._on_adding_ancestor(parent)

    def _on_adding_ancestor(self, ancestor):
        '''called immediately before adding an ancestor to self (i.e., self.parent, parent of self.parent, etc)
        Tracks depth, height, size. This method does not connect any parents or children.
        '''
        self.depth = self.parent.depth + 1
        for child in self.children:
            child._on_adding_ancestor(ancestor)
        
    def _on_added_child(self, child):
        '''called immediately after adding a child to self.
        Tracks depth, height, size. This method does not connect any parents or children.
        '''
        self._on_added_descendant(child)
    
    def _on_added_descendant(self, descendant):
        '''called immediately after adding a descendant to self (or any of self.children).
        Tracks depth, height, size. This method does not connect any parents or children.
        '''
        self.height = 1 + max(child.height for child in self.children)
        self.size += descendant.size
        if self.parent is not None:
            self.parent._on_added_descendant(descendant)

    # # # CHILDREN # # #
    def make_child(self, obj):
        '''makes a child of self, with obj as its stored object, and returns the child.'''
        child = type(self)(obj, parent=self)
        return child

    def make_children(self, objs):
        '''make_child(obj) for obj in objs; returns the list of newly made children.'''
        return [self.make_child(obj) for obj in objs]

    # # # ITERATION # # #
    def flat(self, *, include_self=False):
        '''returns a generator which iterates over all of self's descendants, in depth-first order.
        if include_self, yield self first.
        '''
        if include_self:
            yield self
        for child in self.children:
            yield child
            for descendant in child.flat():
                yield descendant

    def __iter__(self):
        '''iterates over self.children.'''
        return iter(self.children)

    def enumerate_flat(self, *, include_self=False):
        '''returns a generator which iterates over all of self's descendants, in depth-first order,
        yielding (index, node) pairs, such that self[index] == node.
            Note that index will be a tuple with length == node.depth.
        if include_self, yield self first, as: ((), self)).
        '''
        if include_self:
            yield (), self
        for i, child in enumerate(self.children):
            yield (i,), child
            for jtuple, descendant in child.enumerate_flat():
                yield (i,) + jtuple, descendant

    def flat_branches_until(self, branches_until, *, include_self=False):
        '''returns a generator which iterates over all of self's descendants, in depth-first order,
        but stop looking at descendants on a branch as soon as branches_until(node).
        E.g. self.flat_branches_until(lambda node: node.obj==7) will be similar to flat,
            but won't go to any descendants for any node with obj==7.

        if include self, yield self first, and check branches_until(self) before continuing.
        otherwise, never check branches_until(self).
        '''
        if include_self:
            yield self
            if branches_until(self):
                return  # stop iteration
        for child in self.children:
            yield child
            if not branches_until(child):
                for descendant in child.flat_branches_until(branches_until):
                    yield descendant

    # # # INDEXING # # #
    def __getitem__(self, i):
        '''returns self.children[i]. If i is a tuple, index repeatedly, e.g. self[i[0]][i[1]][i[2]]...'''
        if isinstance(i, tuple):
            result = self
            for j in i:
                result = result[j]
            return result
        else:
            return self.children[i]

    # # # DISPLAY # # #
    def __repr__(self):
        contents = [f'[depth={self.depth}, height={self.height}, size={self.size}]']
        contents.append(f'obj={repr_simple(self.obj)}')
        return f'{self.__class__.__name__}({", ".join(contents)})'

    def _shorthand_repr(self):
        '''returns shorthand repr for this node (without children): "((depth, height, size)) obj".'''
        return f'(({self.depth}, {self.height}, {self.size})) {repr_simple(self.obj)}'

    DEFAULT_TREE_SHOW_DEPTH = None
    DEFAULT_TREE_SHOW_MAX_DEPTH = None
    DEFAULT_TREE_SHORTHAND = None

    def html(self, show_depth=None, max_depth=None, *, shorthand=None):
        '''returns html for displaying self and all of self's children.
        show_depth: None or int
            max number of layers of tree to show by default (i.e. "not hidden" by default)
            None --> use self.DEFAULT_TREE_SHOW_DEPTH if defined else DEFAULTS.TREE_SHOW_DEPTH
        max_depth: None or int
            max number of layers of tree to render (even if all layers are "not hidden").
            Anything deeper will not be converted to html string.
            None --> use self.DEFAULT_TREE_SHOW_MAX_DEPTH if defined else DEFAULTS.TREE_SHOW_MAX_DEPTH
        shorthand: None or bool
            whether to use shorthand for the "Tree([depth=N, height=N, size=N], obj=...)" part of the repr.
            True --> use shorthand; replace that^ with: "((N, N, N)) ..."
            None --> use self.DEFAULT_TREE_SHORTHAND if defined else DEFAULTS.TREE_SHORTHAND
        '''
        if show_depth is None: show_depth = getattr(self, 'DEFAULT_TREE_SHOW_DEPTH', None)
        if show_depth is None: show_depth = DEFAULTS.TREE_SHOW_DEPTH
        if max_depth is None: max_depth = getattr(self, 'DEFAULT_TREE_SHOW_MAX_DEPTH', None)
        if max_depth is None: max_depth = DEFAULTS.TREE_SHOW_MAX_DEPTH
        if shorthand is None: shorthand = getattr(self, 'DEFAULT_TREE_SHORTHAND', None)
        if shorthand is None: shorthand = DEFAULTS.TREE_SHORTHAND
        open = (self.height == 0) or (self.depth < show_depth)  # height==0 always open, to reveal no children.
        open_str = ' open' if open else ''
        str_self = self._shorthand_repr() if shorthand else repr(self)
        str_self = str_self.replace('<', '&lt').replace('>', '&gt')  # escape '<' & '>' characters.
        result = f'<details{open_str}><summary>{str_self}</summary>'  # pre prevents html formatting.
        if self.depth < max_depth:
            for child in self.children:
                result += child.html(show_depth=show_depth, max_depth=max_depth, shorthand=shorthand)
        else:
            result += f'... hidden; max render depth={max_depth} ...</p>'
        result += '</details>'
        return result

    def _repr_html_(self):
        '''returns html string for displaying self. (this display hook is used by Jupyter automatically.).

        Includes self.html() and DEFAULTS.TREE_CSS.

        [TODO] apply the css to ONLY this output cell in Jupyter, instead of all output cells...
        '''
        result = f'{DEFAULTS.TREE_CSS}\n{self.html()}'
        return result

    def display(self, show_depth=None, max_depth=None, *, shorthand=None):
        '''display self in html. Includes self.html() and DEFAULTS.TREE_CSS.
        show_depth: None or int
            max number of layers of tree to show by default (i.e. "not hidden" by default)
            None --> use self.DEFAULT_TREE_SHOW_DEPTH if defined else DEFAULTS.TREE_SHOW_DEPTH
        max_depth: None or int
            max number of layers of tree to render (even if all layers are "not hidden").
            Anything deeper will not be converted to html string.
            None --> use self.DEFAULT_TREE_SHOW_MAX_DEPTH if defined else DEFAULTS.TREE_SHOW_MAX_DEPTH
        shorthand: None or bool
            whether to use shorthand for the "Tree([depth=N, height=N, size=N], obj=...)" part of the repr.
            True --> use shorthand; replace that^ with: "((N, N, N)) ..."
            None --> use self.DEFAULT_TREE_SHORTHAND if defined else DEFAULTS.TREE_SHORTHAND
        '''
        s = self.html(show_depth=show_depth, max_depth=max_depth, shorthand=shorthand)
        result = f'{DEFAULTS.TREE_CSS}\n{s}'
        display(HTML(s))



"""
File purpose: tools related to colorbar (see also: colors.py)
"""
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.colorbar as cbar

from ..plot_settings import PlotSettings
from ...tools import format_docstring, UNSET
from ...errors import MappableNotFoundError, PlottingAmbiguityError
from ...defaults import DEFAULTS


_paramdocs_colorbar = {
    'ax': '''ax: None or axes object
        the axes which will inform the size and position for cax.
        None --> use plt.gca()''',
    'mappable': '''mappable: None or mpl.cm.ScalarMappable
        the mappable for this colorbar.
        if None, attempt to find it using find_mappable(ax=ax, im=mappable).''',
    'cax': '''cax: None or axes object
        the axes for this colorbar.
        None --> use make_cax(...), or make cax using mpl_get_cax(...), depending on cax_mode.''',
    'location': f'''location: None, 'right', 'left', 'top', or 'bottom'
        location of colorbar relative to image.
        None --> use DEFAULTS.PLOT.CAX_LOCATION (default: {DEFAULTS.PLOT.CAX_LOCATION})''',
    'sca': '''sca: bool (default False)
        whether to adjust the plt.gca() to cax.
        False --> plt.gca() will be unchanged by this operation.''',
    'ticks_position': '''ticks_position: None (default), 'right', 'left', 'top', or 'bottom'
        None -> ticks are on opposite side of colorbar from image.
        string -> use this value to set ticks position.''',
    'pad': f'''pad: None or number (default 0.01)
        padding between cax and ax, as a percentage of ax size.
        None --> use DEFAULTS.PLOT.CAX_PAD (default: {DEFAULTS.PLOT.CAX_PAD})''',
    'size': f'''size: None or number (default 0.02)
        size of colorbar, as a percentage of ax size.
        None --> use DEFAULTS.PLOT.CAX_SIZE (default: {DEFAULTS.PLOT.CAX_SIZE})''',
}


''' --------------------- making a colorbar - helper functions --------------------- '''

@format_docstring(**_paramdocs_colorbar)
def make_cax(ax=None, location=None, *, sca=False, ticks_position=None,
             pad=None, size=None, **kw_add_axes):
    ''' Creates an axis appropriate for putting a colorbar. Does not "steal" space; makes a new axis.
    This means the colorbar will appear to be the appropriate size even when using aspect='equal',
    however it might seem to have some placement issues if using sharex=True or sharey=True.

    Troubleshooting... if colorbar appears cutoff, off the side of the figure (e.g. if making a movie),
        you might want to use plt.subplots_adjust(right=0.8 (or smaller)),
        to ensure the colorbar actually technically appears within the figure's bbox.
        (Or top=0.8 (or smaller), or left or bottom=0.2 (or larger), depending on cbar location.)

    {ax}
    {location}
        Note: you will still want to set orientation appropriately, via location or orientation.
    {sca}
    {ticks_position}
        Note: you will still want to set colorbar orientation appropriately, via location or orientation.
        NOTE: for horizontal colorbars, the ticks_position may be overriden by the plt.colorbar() call.
    {pad}
    {size}

    additional kwargs go to plt.gcf().add_axes(...)

    Adapted from https://stackoverflow.com/a/56900830.
    Returns cax.
    '''
    if pad is None: pad = DEFAULTS.PLOT.CAX_PAD
    if size is None: size = DEFAULTS.PLOT.CAX_SIZE
    if location is None: location = DEFAULTS.PLOT.CAX_LOCATION
    if ax is None:
        ax = plt.gca()
    p = ax.get_position()
    # calculate cax params.
    ## fig.add_axes(rect) has rect=[x, y, w, h],
    ## where x and y are location for lower left corner of axes.
    ## and w and h are width and height, respectively.
    assert location in ('right', 'left', 'top', 'bottom')
    if location in ('right', 'left'):
        y = p.y0
        h = p.height
        w = size
        if location == 'right':
            x = p.x1 + pad
        else: #'left'
            x = p.x0 - pad - w
    else: #'top' or 'bottom'
        x = p.x0
        w = p.width
        h = size
        if location == 'top':
            y = p.y1 + pad
        else: #'bottom'
            y = p.y0 - pad - h

    # make the axes
    ax0 = plt.gca()
    try:
        cax = plt.gcf().add_axes([x, y, w, h], **kw_add_axes)
    finally:
        if not sca:  # (don't set sca to colorbar, i.e., need to restore ax0.)
            plt.sca(ax0)
    
    # Change ticks position
    if ticks_position is None:
        ticks_position = location
    if ticks_position in ('left', 'right'):
        cax.yaxis.set_ticks_position(ticks_position)
    else: #'top' or 'bottom'
        cax.xaxis.set_ticks_position(ticks_position)

    return cax

def find_mappable(ax=None, *, im=None):
    '''return the relevant mappable. By default, return plt.gci().
    if plt.gci() is None, instead attempt to find a mappable on plt.gca();
        if there is exactly 1, return it, else raise PlottingError.
        the error will be MappableNotFoundError if 0, else PlottingAmbiguityError.

    providing im uses im instead of plt.gci()
    providing ax using ax instead of plt.gca()  (only if didn't provide im)
    
    ax: None or axes object
        if provided, use ax.findobj to find mappable.
        (cannot provide both ax and im)
        if None, check plt.gci() first, then use findobj if that is also None.
    im: None or mappable.
        if provided, return it.
        if None, attempt plt.gci(); returning it if that is not None
    '''
    if (ax is None) and (im is None):
        im = plt.gci()
    elif (ax is not None) and (im is not None):
        raise ValueError('cannot provide both ax and im.')
    if im is not None:
        return im
    # else, im and plt.gci() are both None, so attempt to find mappable on ax.
    if ax is None:
        _ax_used = 'plt.gca()'  # for more informative error string, if necessary
        ax = plt.gca()
    else:
        _ax_used = 'ax provided'  # for more informative error string, if necessary
    search = ax.findobj(lambda obj: isinstance(obj, mpl.cm.ScalarMappable))
    if len(search)==0:
        raise MappableNotFoundError(f'plt.gci()=None, and no mappable found on {_ax_used}.')
    elif len(search)>1:
        raise PlottingAmbiguityError(f'multiple mappables found on axes: {search}')
    else:
        return search[0]

def mpl_get_cax(mappable, *, cax=None, ax=None, use_gridspec=True, **kwargs):
    '''get the cax to use for a colorbar, using the same logic as plt.colorbar().
    (This is adapted from copy-pasting the source code from plt.colorbar().'''
    if cax is not None:
        return cax
    if ax is None:
        ax = getattr(mappable, "axes", None)
    if ax is None:
        raise ValueError(
            'Unable to determine Axes to steal space for Colorbar. '
            'Either provide the *cax* argument to use as the Axes for '
            'the Colorbar, provide the *ax* argument to steal space '
            'from it, or add *mappable* to an Axes.')
    fig = (  # Figure of first axes; logic copied from make_axes.
        [*ax.flat] if isinstance(ax, np.ndarray)
        else [*ax] if np.iterable(ax)
        else [ax])[0].figure
    current_ax = fig.gca()
    if (fig.get_layout_engine() is not None and
            not fig.get_layout_engine().colorbar_gridspec):
        use_gridspec = False
    if (use_gridspec
            and isinstance(ax, mpl.axes._base._AxesBase)
            and ax.get_subplotspec()):
        cax, kwargs = cbar.make_axes_gridspec(ax, **kwargs)
    else:
        cax, kwargs = cbar.make_axes(ax, **kwargs)
    # make_axes calls add_{axes,subplot} which changes gca; undo that.
    fig.sca(current_ax)
    cax.grid(visible=False, which='both', axis='both')
    return cax

''' --------------------- Colorbar and colorbar --------------------- '''
# Colorbar is the colorbar class here (like Colorbar from matplotlib but with more convenience);
# colorbar is like plt.colorbar but creates an instance of this Colorbar class.

class Colorbar(mpl.colorbar.Colorbar):
    '''matplotlib.colorbar.Colorbar class with more methods.
    use Colorbar.here(...) to create a colorbar if cax and/or mappable not known.
    '''
    # # # CREATION / INITIALIZATION # # #
    def __init__(self, cax, mappable, *args_super, **kw_super):
        super().__init__(cax, mappable, *args_super, **kw_super)

    find_mappable = staticmethod(find_mappable)

    @classmethod
    @PlotSettings.format_docstring(**_paramdocs_colorbar, sub_indent=DEFAULTS.TAB)
    def here(cls, mappable=None, *, cax=None, ax=None, cax_mode=UNSET,
             location=None, sca=False, ticks_position=None,
             pad=None, size=None, kw_add_axes=dict(), **kw_colorbar):
        '''create a colorbar, like plt.colorbar(), but using Colorbar class instead.
        {mappable}
        {cax}
            kwargs for make_cax would be:
                ax, location, sca, ticks_position, pad, size, **kw_add_axes.
        cax_mode: {cax_mode}
        {location}
            also passed to super().__init__, so that orientation is set appropriately.

        The following are relevant if cax not provided, and using cax_mode='mpl':
        {sca}
        {ticks_position}
        {pad}
        {size}
        kw_add_axes: dict
            passed to make_cax(..., **kw_add_axes)... which passes it to plt.gcf().add_axes().

        additional kwargs get passed to matplotlib.colorbar.Colorbar.__init__.
        '''
        if mappable is None:
            mappable = cls.find_mappable(ax=ax, im=mappable)
        if location is None:
            location = DEFAULTS.PLOT.CAX_LOCATION
        if cax is None:
            cax_mode = PlotSettings(cax_mode=cax_mode).get('cax_mode')
            if cax_mode == 'pc':
                cax = make_cax(ax=ax, location=location, sca=sca, ticks_position=ticks_position,
                               pad=pad, size=size, **kw_add_axes)
            elif cax_mode == 'mpl':
                # get the relevant kwargs; PlotSettings knows how to do it, so we can just pass all the kwargs.
                kw = PlotSettings.cls_get_mpl_kwargs('plt.colorbar',
                            cax=cax, ax=ax, cax_mode=cax_mode, location=location,
                            sca=sca, ticks_position=ticks_position, pad=pad, size=size,
                            kw_add_axes=kw_add_axes, **kw_colorbar)
                # PlotSettings doesn't know how to handle converting None to rcParams defaults..
                if 'pad' in kw and kw['pad'] is None: del kw['pad']
                # actually get the cax:
                cax = mpl_get_cax(mappable, **kw)
        return cls(cax, mappable, **kw_colorbar)

    # # # MISC. PROPERTIES # # #
    label = property(lambda self: self.get_label(),
                     lambda self, value: self.set_label(value),
                     doc='''the label on this colorbar. Works regardless of orientation.''')
    def get_label(self):
        '''return the label on this colorbar.'''
        return self.ax.get_ylabel() if (self.orientation=='vertical') else self.ax.get_xlabel()


@format_docstring(doc_here=Colorbar.here.__doc__)
def colorbar(mappable=None, *, cax=None, ax=None,
             location=None, size=None, pad=None, **kw_colorbar_here):
    '''{doc_here}'''
    __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
    return Colorbar.here(mappable=mappable, cax=cax, ax=ax,
                         location=location, size=size, pad=pad,
                         **kw_colorbar_here)



"""
File Purpose: The main single-fluid mhd calculator class, for hookups to inherit from.
"""

from .mhd_bases import MhdBasesLoader
from .mhd_eos_loader import MhdEosLoader
from ..plasma_calculator import PlasmaCalculator

class MhdCalculator(MhdEosLoader, MhdBasesLoader, PlasmaCalculator):
    '''class for single-fluid mhd plasma calculator object.

    Not intended for direct instantiation. Instead, see options in the "hookups" subpackage,
        or write your own hookup for a different type of input, following the examples there.
    '''
    pass



"""
File Purpose: testing multiprocessing
"""

import pickle

import numpy as np
import pytest

import PlasmaCalcs as pc

MAX_TIMEOUT = 15  # timeout after this many seconds for each test below;
# in case something goes horribly wrong! (don't want to waste gitlab CI time for no reason...)

''' --------------------- multiprocessing generic --------------------- '''

def test_multiprocessing_basics():
    '''test that basic multiprocessing works.'''
    for ncpu in [1,2]:  # (when ncpu=1, will not use multiprocessing. When ncpu=2, will use it.)
        # ensure that timeout works:
        with pytest.raises(pc.TimeoutError):
            tasks = pc.TaskList((pc.mptest_sleep, 2), (pc.mptest_sleep, 2))
            tasks(ncpu=ncpu, timeout=1)

        # ensure that results are correct:
        tasks = pc.TaskList([pc.mptest_add100, 1], [pc.mptest_add100, 2], [pc.mptest_add100, 3])
        results = tasks(ncpu=ncpu, timeout=MAX_TIMEOUT)
        assert results == [101, 102, 103]

        # ensure that it works even with lots of tasks:
        tasks = pc.TaskList(*[(pc.mptest_add100, i) for i in range(100)])
        results = tasks(ncpu=ncpu, timeout=MAX_TIMEOUT)
        assert results == [i+100 for i in range(100)]

 # [TODO] test TaskArray

def test_multiprocessing_being_used():
    '''test that multiprocessing is ACTUALLY being used.'''
    # if the tasks are not actually being run with multiprocessing, this test will fail;
    # sleeping for 4 seconds, twice in a row, in less than 6 seconds total, is impossible!!!
    # but if the sleeps are run in parallel, then it will be possible :)
    #
    # (this test assumes that the multiprocessing overhead here will take less than 2 seconds;
    #  if the overhead takes longer than 2 seconds, this test will fail erroneously.)
    with pc.TimeLimit(6):
        tasks = pc.TaskList([pc.mptest_sleep, 4], [pc.mptest_sleep, 4])
        # ensure that multiprocessing is ACTUALLY being used:
        results = tasks(ncpu=2, timeout=MAX_TIMEOUT)


''' --------------------- multiprocessing + sentinels --------------------- '''

def assert_is_unset(unset):
    '''asserts that unset is pc.UNSET.
    Defined as a top-level function so it can be compatible with multiprocessing.
    '''
    assert unset is pc.UNSET

def test_basic_sentinels():
    '''test that basic sentinels work with multiprocessing.'''
    timeout = 10  # timeout after 10 seconds. In case something goes horribly wrong.
    tasks = pc.TaskList((assert_is_unset, pc.UNSET), (assert_is_unset, pc.UNSET))
    tasks(ncpu=2, timeout=timeout)

    tasks_many = pc.TaskList(*[(assert_is_unset, pc.UNSET) for _ in range(100)])
    tasks_many(ncpu=2, timeout=timeout)

def assert_is_missing_snap(missing_snap):
    '''asserts that missing_snap is pc.MISSING_SNAP.
    Defined as a top-level function so it can be compatible with multiprocessing.
    '''
    assert missing_snap is pc.MISSING_SNAP

def test_dimension_value_sentinels():
    '''test that UniqueDimensionValue objects work with multiprocessing.'''
    timeout = 10  # timeout after 10 seconds. In case something goes horribly wrong.
    tasks = pc.TaskList((assert_is_missing_snap, pc.MISSING_SNAP), (assert_is_missing_snap, pc.MISSING_SNAP))
    tasks(ncpu=2, timeout=timeout)


''' --------------------- picklable PlasmaCalculator --------------------- '''

def test_picklable_plasma_calculator():
    '''test that PlasmaCalculator instance can be pickled (i.e. is compatible with multiprocessing)'''
    p = pc.PlasmaCalculator()
    # immediately after initialization
    dump = pickle.dumps(p)
    _tmp = pickle.loads(dump)
    # after getting a var
    p('0')
    dump = pickle.dumps(p)
    _tmp = pickle.loads(dump)



"""
File Purpose: units in mhd
"""
import numpy as np

from ..defaults import DEFAULTS
from ..units import UnitsManager, _units_manager_helpstr
from ..tools import format_docstring, UNSET


''' --------------------- MhdUnitsManager --------------------- '''

@format_docstring(helpstr=_units_manager_helpstr, sub_ntab=1)
class MhdUnitsManager(UnitsManager):
    '''units manager with the from_mhd method,
    which determines all units based on u_l, u_t, u_r, and mu0_raw.

    Note: cgs units for electromagnetic quantities are not supported here,
        because cgs electromagnetic equations differ from SI equations,
        e.g. cgs includes extra factors of 4*pi or c in various places.

    How to infer electromagnetic units from t, l, r, and mu0?
        Internally, this method determines:
            u_M = u_r * u_l**3   # mass
            u_N = u_M * u_l * u_t**-2   # energy (Newtons in SI)
        Amperes units:
            (mu0 [si]) = u_N * u_A**-2 * (mu0 [raw])
            --> u_A = sqrt(u_N * (mu0 [raw]) / (mu0 [si]))
            where mu0 [si] == DEFAULTS.PHYSICAL.CONSTANTS_SI['mu0'], approx. 4 * pi * 1e-7.
        Which yields, for charge units:
            u_q = u_A * u_t
        At which point, we have u_l, u_t, u_M, and u_q,
            which is sufficient to infer all other units (aside from K for temperature)

    --- MhdUnitsManager.help() will print a helpful message: ---
        {helpstr}
    '''
    @classmethod
    def from_mhd(cls, units='si', *, u_l, u_t, u_r, mu0_raw=1, K=1):
        '''create a MhdUnitsManager from u_l, u_t and u_r, the SI conversion factors.
        CAUTION: these are the SI conversion factors. NOT cgs.
        (some MHD simulations, like Bifrost, use similar names for the CGS factors.)
        
        units: 'si' or 'raw'
            by default, outputs of the resulting UnitsManager convert to this unit system.
            Can easily change later by setting result.units to a different value.
        u_l: number
            length [si] = u_l * length [raw]
        u_t: number
            time [si] = u_l * time [raw]
        u_r: number
            mass density [si] = u_r * mass density [raw]
        mu0_raw: number, default 1
            value of mu0 in 'raw' units system. Used to infer Amperes units via:
                u_M = u_r * u_l**3          # mass
                u_N = u_M * u_l * u_t**-2   # energy (Newtons in SI)
                mu0_si = u_N * u_A**-2 * mu0_raw
                u_A = sqrt(u_N * mu0_raw / mu0_si)
            where mu0_si == DEFAULTS.PHYSICAL.CONSTANTS_SI['mu0'], approx. 4 * pi * 1e-7.
        K: number, default 1
            temperature [si] = K * temperature [raw]
        '''
        u_M = u_r * u_l**3
        u_N = u_M * u_l * u_t**-2
        u_A = np.sqrt(u_N * mu0_raw / DEFAULTS.PHYSICAL.CONSTANTS_SI['mu0'])
        u_q = u_A * u_t
        return cls(units=units, q=u_q, M=u_M, l=u_l, t=u_t, K=K)

    @classmethod
    def from_mhd_cgs(cls, units='si', *, ucgs_l, ucgs_t, ucgs_r, mu0_raw=1, K=1):
        '''create a BifrostUnitsManager from ucgs_l, ucgs_t and ucgs_r, the CGS conversion factors.

        units: 'si' or 'raw'
            by default, outputs of the resulting UnitsManager convert to this unit system.
            Can easily change later by setting result.units to a different value.
        ucgs_l: number
            length [cgs] = ucgs_l * length [raw]
        ucgs_t: number
            time [cgs] = ucgs_t * time [raw]
        ucgs_r: number
            mass density [cgs] = ucgs_r * mass density [raw]
        mu0_raw: number, default 1
            value of mu0 in 'raw' units system. Used to infer Amperes units via:
                u_M = u_r * u_l**3          # mass
                u_N = u_M * u_l * u_t**-2   # energy (Newtons in SI)
                mu0_si = u_N * u_A**-2 * mu0_raw
                u_A = sqrt(u_N * mu0_raw / mu0_si)
            where mu0_si == DEFAULTS.PHYSICAL.CONSTANTS_SI['mu0'], approx. 4 * pi * 1e-7.
        K: number, default 1
            temperature [cgs] = K * temperature [raw]
        '''
        # length [si]  = u_l    * length [raw]
        # length [cgs] = ucgs_l * length [raw]
        # --> u_l / ucgs_l = length [si] / length [cgs]
        # e.g. length [si] = 1, length [cgs] = 100 --> u_l = ucgs_l * 1e-2
        u_l = ucgs_l * 1e-2
        # s to s... cgs and si use same time units
        u_t = ucgs_t
        # similar logic as for length... but now we have g/cm^3 to kg/m^3,
        # so g to kg gives 10^-3, and cm^-3 to m^-3 gives 10^6
        u_r = ucgs_r * 1e3
        return cls.from_mhd(units=units, u_l=u_l, u_t=u_t, u_r=u_r, mu0_raw=mu0_raw, K=K)

    # # # DISPLAY # # #
    def _repr_show_factors(self):
        '''returns dict of name and conversion factor to si, to include in repr(self).
        Here, include l, t, and r. Also include K if it is not 1.
        Also include mu0_raw if it is not 1.
        '''
        factors = {key: self(key, 'si', 'raw') for key in ('l', 't', 'r', 'K')}
        if factors['K'] == 1:
            del factors['K']
        if not np.isclose(self('mu0', 'raw'), 1):
            factors['mu0_raw'] = self('mu0', 'raw')
        return factors



"""
Package Purpose: handling units
"""
from .units_haver import UnitsHaver, CoordsUnitsHaver
from .units_manager import (
    UnitsManager,
    units_string_to_si_bases, UnitsStringManager,
    _units_manager_helpstr,
)
from .units_manager_pic import UnitsManagerPIC



import math
import os

import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

import PlasmaCalcs

# INPUT PARAMETERS
path_to_folder = r"G:\My Drive\Research\eppic_out\eregion\lowerE"
fluid = 1
plot_quantity = "n"
plot_every = 11
n_columns = 2 # number of plot columns
slices = {"y": 0}
pdf_name = "den1_xz"

# LOAD DATA
ec = PlasmaCalcs.EppicCalculator.from_here(os.path.join(path_to_folder, "eppic.i"), kw_units={"M": 1}, u_l=1)
ec.fluid = fluid
sample_snaps = ec.snaps[::plot_every]

# CREATE CANVAS
n_rows = math.ceil(len(sample_snaps) / n_columns)
fig, axs = plt.subplots(nrows=n_rows, ncols=n_columns, squeeze=False, figsize=[8.5, 11])

# SET COUNTERS
this_row = 0
this_column = 0

# SET PDF PATH
pdf_path = os.path.join(path_to_folder, f"{pdf_name}.pdf")

with PdfPages(pdf_path) as pdf:
    # LOOP THROUGH SNAPS
    for snap in sample_snaps:
        time = snap.t
        ec.snap = snap
        # PLOT
        plot_title = f"Density {fluid} at t = {time}."
        this_axis = axs[this_row, this_column]
        array = ec(plot_quantity).isel(**slices)
        array.plot(cmap="plasma", ax=this_axis, rasterized=True)
        this_axis.set_title(plot_title)
        # UPDATE ROW & COLUMN
        if this_column == n_columns - 1:
            this_row += 1
            this_column = 0
        else:
            this_column += 1

    # MAKE SURE PLOTS DON'T OVERLAP
    plt.tight_layout()

    # SAVE TO PDF
    pdf.savefig()
    plt.close()
    print(f"File saved to {pdf_path}.")




"""
File Purpose: class to help with managing making plots.
[TODO] allow multiple people to have same-named plotting function.
[TODO] allow plotter patterns, e.g. 'subN' to do a movie but subsample to every Nth snapshot.
"""
import contextlib
import os

import matplotlib.pyplot as plt

from ..defaults import DEFAULTS
from ..errors import (
    InputError, InputConflictError, InputMissingError,
    PlottingNotImplementedError, PlottingAmbiguityError,
)
from ..quantities.quantity_loader import MetaQuantTracking
# ^ specified .quantity_loader to help avoid circular imports.
from ..tools import (
    format_docstring, UNSET,
    ProgressUpdater, code_snapshot_info,
)


''' --------------------- Plotter --------------------- '''

_paramdocs = {
    'fname': '''str
        name of the callable used to make the plot,
        via getattr(obj,fname)(...), where obj is a PlotterManager.''',
    'who': '''str or list of str
        "person" or people associated with this plotter. Arbitrary strings.
        E.g. to only make plots for 'sam' person, use PlotterManager.standard_plots(..., who='sam').''',
    'kind': '''UNSET, str, or list of str
        "kind" or kinds associated with this plotter. Arbitrary strings.
        E.g. to only make plots for 'movie' kind, use PlotterManager.standard_plots(..., kind='movie').
        Internally, stored at Plotter.kinds (plural).''',
    'kinds': '''UNSET, str, or list of str
        alias for kwarg `kind`. Can provide `kinds` or `kind` but not both.''',
    'ani': '''UNSET or bool
        whether the plotter is for an animated plot.
        (commands to save/show an animated plot are different than those to save/show a static plot.)
        if UNSET, infer from kinds. True if kinds includes any of these (case insensitive):
            ['movie', 'gif', 'ani', 'animation', 'animated']''',
    'savename': '''None or str
        default filename when doing plotter.save() for this plot.
        None --> use plotter.name, which is fname after removing 'plot_' prefix.''',
    'cost': '''number
        some guess about the cost to make this plot. Default=20.
        when making multiple plots, go in cost order.
        Can also set cost thresholds and skip ones that are too expensive.''',
    'aliases': '''list of str
        self.known[(person, alias)] = self.known[(person, name)] for each alias provided.''',
}

_paramdocs_plot = {
    'save': '''bool, str, or dict.
        whether to save figure after calling plotter.
        str --> filename=save.format(name=name, savename=savename), instead of default filename=name.
        dict --> pass to saver as kwargs. Use kwarg 'dst' to also provide filename.
                if plotter.ani, saver is movie_obj.ani(), else saver is plt.savefig().''',
    'show': '''bool
        whether to plt.show() after making plot (and, after save).
        if show when ani==True, return movie_obj.ani() (so it will display in jupyter)''',
    'close': '''bool
        whether to plt.close() after making plot (and, after save/show).
        (ignored when show==True and ani==True)''',
}

@format_docstring(**_paramdocs)
class Plotter():
    '''plotter which can be used by a PlotterManager.
    PlotterManager.KNOWN_PLOTTERS will be a dict of {{str: Plotter}} pairs.

    fname: {fname}
    f: None or callable
        the function to call (with PlotterManager as first arg) to make the plot.
        None --> get when needed, via getattr(plot_manager, plotter.fname).
    who: {who}
    kind: {kind}
    kinds: {kinds}
    ani: {ani}
    savename: {savename}
    cost: {cost}
    '''
    MOVIE_KINDS = ['movie', 'gif', 'ani', 'animation', 'animated']
    DEFAULT_COST = 20

    def __init__(self, fname, *, f=None, who=[], kind=UNSET, kinds=UNSET, ani=UNSET,
                 savename=None, cost=DEFAULT_COST, **kw_super):
        self.fname = fname
        self.f = f
        self.who = [who] if isinstance(who, str) else who
        if kind is not UNSET and kinds is not UNSET:
            raise InputConflictError(f'cannot provide both kind and kinds. Got kind={kind!r}, kinds={kinds!r}')
        if kinds is UNSET: kinds = kind
        if kinds is UNSET: kinds = []  # if still UNSET, then both kind and kinds were UNSET.
        self.kinds = [kinds] if isinstance(kinds, str) else kinds
        if ani is UNSET: ani = self._kinds_imply_ani()
        self.ani = ani
        self.savename = savename
        self.cost = cost
        super().__init__(**kw_super)

    @property
    def name(self):
        '''name of this plotter, assuming self.name looks like 'plot_name'.'''
        START = 'plot_'
        fname = self.fname
        assert fname.startswith(START), f'f.__name__ must start with {START!r}; got {fname!r}'
        return fname[len(START):]

    @property
    def savename(self):
        '''name of file to save this plot to, if not provided explicitly.
        self.savename = None --> use self.name.
        self.savename = other value --> use that value.
        '''
        return self.name if self._savename is None else self._savename
    @savename.setter
    def savename(self, value):
        self._savename = value

    # # # DETERMINE IF ANIMATED # # #
    def _kinds_imply_ani(self):
        '''infers value for self.ani, based on self.kinds.'''
        return any(k.lower() in self.MOVIE_KINDS for k in self.kinds)

    # # # PLOTTING # # #
    @format_docstring(**_paramdocs_plot, sub_ntab=1)
    def plot(self, plot_manager, *,
             save=False, show=False, close=False,
             **kw_plotter):
        '''actually call the plotter function (plot_manager.f) associated with this Plotter.
        returns plotter(**kw_plotter) (unless show=True and ani=True. Then return movie_obj.ani()).

        Might also save, show, and/or close the plot, depending on kwargs.

        save: {save}
        show: {show}
        close: {show}
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        # do the plotting
        if self.f is None:
            plotter = getattr(plot_manager, self.fname)
            result = plotter(**kw_plotter)
        else:
            result = self.f(plot_manager, **kw_plotter)
        if save != False:
            if save == True: save = dict()
            elif isinstance(save, str): save = dict(dst=save)
            else: save = save.copy()  # don't alter the user-input dict directly
            self.save(movie_obj=result, **save)
        if show:
            shown = self.show(movie_obj=result)
            if self.ani:
                return shown
        if close:
            plt.close()
        return result

    def save(self, movie_obj=None, dst=None, *args_save, bbox_inches=UNSET, **kw_save):
        '''saves the current figure. Default behaviors appropriate if current figure is self.plot().
        does movie_obj.save(...) if self.ani, else plt.savefig(...).
        returns abspath to the created file.
        
        movie_obj: None or object
            if self.ani, must provide movie_obj; will use movie_obj.ani(fname, *args, **kw)
        dst: None or str
            file name for where to save the figure. os.makedirs(exist_ok=True) as needed.
            None --> use self.savename
            str --> use dst.format(name=self.name, savename=self.savename)
        bbox_inches: UNSET or any value
            if provided, pass to plt.savefig(...) but NOT movie_obj.save(...)
        additional args and kwargs go to ani() or savefig().
        '''
        if self.ani and movie_obj is None:
            raise InputMissingError('must provide movie_obj if self.ani')
        if dst is None:
            dst = self.savename
        else:
            dst = dst.format(name=self.name, savename=self.savename)
        dst = os.path.abspath(dst)
        os.makedirs(os.path.dirname(dst), exist_ok=True)
        if self.ani:
            movie_obj.save(dst, *args_save, **kw_save)
        else:
            if bbox_inches is not UNSET: kw_save['bbox_inches'] = bbox_inches
            plt.savefig(dst, *args_save, **kw_save)
        return dst

    def show(self, movie_obj=None, **kw_ani):
        '''shows the current figure. Default behaviors appropriate if current figure is self.plot().
        does movie_obj.ani() if self.ani, else plt.show().
        additional kwargs go to ani() if calling ani().
        '''
        if self.ani and movie_obj is None:
            raise InputMissingError('must provide movie_obj if self.ani')
        if self.ani:
            return movie_obj.ani(**kw_ani)
        else:
            return plt.show()

    # # # HASHING # # #
    def __hash__(self):
        return hash((type(self), self.fname, self.f, tuple(self.who), tuple(self.kinds), self.ani, self.savename))

    # # # DISPLAY # # #
    def _repr_contents(self):
        '''return contents for __repr__'''
        result = [f'{self.fname!r}', f'who={self.who!r}', f'kinds={self.kinds!r}']
        if self.ani != self._kinds_imply_ani():
            result.append(f'ani={self.ani!r}')
        if self.cost != 20:
            result.append(f'cost={self.cost!r}')
        # f_module = self.f_module()
        # if f_module is not None:
        #     result.append(f'f_module={f_module!r}')
        return result
    
    def __repr__(self):
        return f'{type(self).__name__}({", ".join(self._repr_contents())})'


class DecoratingPlotters():
    '''Decorating functions to put plotters into a dict (self.known)

    known: None or dict
        will be a dict of {(name, person): Plotter} pairs.
        person will be None if it nobody was specified via `who`.

    self.unique will be a dict of {id(plotter): plotter} for all unique plotters.
    self.known_names will be a set of all names appearing in any plotter in self.known.
    self.known_kinds will be a set of all kinds appearing in any plotter in self.known.
    self.known_who will be a set of all people appearing in any plotter in self.known.

    Example: known_plotter = DecoratingPlotters(storage_dict)
        @known_plotter
        def plot_name1(...):
            ...
        @known_plotter(who='sam', kinds=['movie', 'density'])
        def plot_name2(...):
            ...
        # at this point, we have:
        #   known_plotter.known == {'name1': Plotter('plot_name1'),
        #                           'name2': Plotter('plot_name2', who='sam', kinds=['movie', 'density'])}
    '''
    plotter_cls = Plotter

    def __init__(self, known=None):
        if known is None: known = dict()
        self.known = known
        self._update_unique()

    def _update_unique(self):
        unique = {}
        for plotter in self.known.values():
            if id(plotter) not in unique:
                unique[id(plotter)] = plotter
        self.unique = unique

    @property
    def known_names(self):
        '''set of all names appearing in any plotter in self'''
        return {p.name for p in self.unique.values()}

    @property
    def known_kinds(self):
        '''set of all kinds appearing in any plotter in self'''
        return {k for p in self.unique.values() for k in p.kinds}

    @property
    def known_who(self):
        '''set of all people appearing in any plotter in self.'''
        return {w for p in self.unique.values() for w in p.who}

    @format_docstring(**_paramdocs, sub_ntab=1)
    def track_f(self, f, *,  who=[], kind=UNSET, kinds=UNSET, ani=UNSET, savename=None,
                cost=Plotter.DEFAULT_COST, aliases=[]):
        '''add f to self.known as a Plotter.
        f must look like 'plot_name' where name can be any string;
            will make self.known[(name, person)] = Plotter(f, ...)

        who: {who}
        kind: {kind}
        kinds: {kinds}
        ani: {ani}
        savename: {savename}
        cost: {cost}
        aliases: {aliases}
        '''
        fname = f.__name__
        START = 'plot_'
        assert fname.startswith(START), f'f.__name__ must start with {START!r}; got {fname!r}'
        name = fname[len(START):]
        plotter = self.plotter_cls(fname, f=f, who=who, kind=kind, kinds=kinds, ani=ani, savename=savename, cost=cost)
        if getattr(self, 'cls_associated_with', None) is not None:
            plotter.cls_where_defined = self.cls_associated_with
        for key in self._get_keys(name, who, aliases):
            self.known[key] = plotter
        # [EFF] might be inefficient to run _update_unique() every time,
        #  but this only occurs when defining new functions to track,
        #  not while running code, so hopefully it's not a big deal.
        self._update_unique()

    def _get_keys(self, name, who, aliases):
        '''get keys for where to store the plotter with this name, in self.known.'''
        if isinstance(who, str):
            who = [who]
        elif len(who)==0:
            who = [None]
        allnames = [name] + list(aliases)
        return [(n, w) for n in allnames for w in who]

    @format_docstring(**_paramdocs, sub_ntab=1)
    def decorator(self, *, who=[], kind=UNSET, kinds=UNSET, ani=UNSET, savename=None,
                  cost=Plotter.DEFAULT_COST, aliases=[]):
        '''returns decorator for plot manager's f(self, *args, **kw), which returns f, unchanged...
        but also sets self.known[name] = Plotter(f.__name__, ...).
        (requires f.__name__ looks like 'plot_name')

        who: {who}
        kind: {kind}
        kinds: {kinds}
        ani: {ani}
        savename: {savename}
        cost: {cost}
        aliases: {aliases}
        '''
        def return_f_after_some_bookkeeping(f):
            '''returns f, unchanged, after calling self.track_f(f, ...) to add it to self.known.'''
            self.track_f(f, who=who, kind=kind, kinds=kinds, ani=ani, savename=savename, cost=cost, aliases=aliases)
            return f
        return return_f_after_some_bookkeeping

    def __call__(self, f=None, *, who=[], kind=UNSET, kinds=UNSET, ani=UNSET,
                 savename=None, cost=Plotter.DEFAULT_COST, aliases=[], **kw_decorator):
        '''if f is provided, return self.decorator(**kw)(f). Otherwise, return self.decorator(**kw).
        This enables instances to be used as decorators directly, i.e. "@self",
            or used as decorators after providing kwargs, e.g. "@self(who='sam', kind='movie')".
        See help(type(self)) for examples.
        See help(self.decorator) for parameter descriptions.
        '''
        kw = dict(who=who, kind=kind, kinds=kinds, ani=ani, savename=savename, cost=cost,
                  aliases=aliases, **kw_decorator)
        if f is None:
            return self.decorator(**kw)
        else:
            return self.decorator(**kw)(f)

    # # # DISPLAY # # #
    def __repr__(self):
        contents = [f'known_names={self.known_names}',
                    f'known_who={self.known_who}',
                    f'known_kinds={self.known_kinds!r}']
        contents_str = ',\n'.join(contents)
        return f'{type(self).__name__}({contents_str})'


''' --------------------- MetaPlotterManager --------------------- '''

class MetaPlotterManager(MetaQuantTracking):
    '''metclass which predefines some things for registering plotters, in the class namespace:
        KNOWN_PLOTTERS - dict of {(var, who): Plotter} for all f decorated with @known_plotter
        known_plotter - use @known_plotter to decorate functions.
        UNIQUE_PLOTTERS - list of unique Plotter objects in this class
                (KNOWN_PLOTTERS.values() lists any Plotter with len(who)>=2 multiple times;
                UNIQUE_PLOTTERS includes each Plotter only once.)
    '''
    # note: inheriting from MetaQuantTracking is required in order to be compatible with QuantityLoader;
    #  the metaclass of a derived class must be a (non-strict) subclass of the metaclasses of all its bases.
    @classmethod
    def __prepare__(_metacls, _name, bases):
        super_result = super().__prepare__(_name, bases)
        KNOWN_PLOTTERS = dict()
        for base in bases:
            # So that keys appear in __mro__ order, must iterate through bases in order.
            # So, the first time a key appears, it points to the earliest (in __mro__ order)
            # implementation, so it should not be overwritten if it appears again.
            #    Hence, avoid dict.update; use dict.setdefault instead.
            #
            # [TODO] double-check these docs. Adapted directly from MetaQuantTracking.
            # Additionally, need to avoid the following:
            #    'vv' defined in Class0.
            #    Class1 subclasses Class0 but does NOT also define 'vv'.
            #    Class2 subclasses Class0 and defines 'vv' too.
            #    class Class3(Class1, Class2): ...
            #    now, Class3.mro() has [Class1, Class2, Class0], so:
            #      Class3.plot_vv uses Class2.plot_vv.
            #        (This is good; should use the earliest-in-mro() plot_vv to get value.)
            #      Class3.KNOWN_PLOTTERS['vv'] uses Class1.KNOWN_PLOTTERS['vv'].
            #        (This is bad; should use the KNOWN_PLOTTERS attached to plot_vv.)
            # To avoid this, below checks where key was defined if it appears in dict already;
            #    if it was defined in a subclass, don't overwrite.
            base_kplots = getattr(base, 'KNOWN_PLOTTERS', {})
            for k, v in base_kplots.items():
                try:
                    existing = KNOWN_PLOTTERS[k]
                except KeyError:  # didn't exist yet
                    KNOWN_PLOTTERS[k] = v
                else:  # existed already
                    # only overwrite if this base is a subclass of where the existing value was defined.
                    if issubclass(base, existing.cls_where_defined):
                        KNOWN_PLOTTERS[k] = v
        known_plotter = DecoratingPlotters(KNOWN_PLOTTERS)
        return dict(**super_result,
                    KNOWN_PLOTTERS=KNOWN_PLOTTERS,
                    known_plotter=known_plotter)

    def __init__(cls, *args, **kw):
        cls.known_plotter.cls_associated_with = cls
        for k, v in cls.KNOWN_PLOTTERS.items():
            if not hasattr(v, 'cls_where_defined'):
                v.cls_where_defined = cls
        unique = {}
        for plotter in cls.KNOWN_PLOTTERS.values():
            if id(plotter) not in unique:
                unique[id(plotter)] = plotter
        cls.UNIQUE_PLOTTERS = list(unique.values())
        super().__init__(*args, **kw)


''' --------------------- PlotterManager --------------------- '''

_paramdocs_get_plotters = {
    'name': '''None or str
        plotter name. E.g. 'deltafrac_n'.
        None --> include all plotters regardless of 'name'.''',
    'who': '''UNSET, None, str, or list.
        person associated with the plotter.
        UNSET --> include plotters regardless of 'who'.
        None --> require len(plotter.who) == 0.
        str --> require this name to be in plotter.who.
        list --> require at least one of these to be in plotter.who
                (or, if None in list, allow len(plotter.who)==0, too)''',
    'kind': '''UNSET, str, or list.
        kind associated with the plotter.
        UNSET --> include plotters regardless of 'kind'.
        str --> require this kind to be in plotter.kinds.
        list --> require at least one of these to be in plotter.kinds.''',
    'all_whos': '''UNSET or list.
        include only plotters with ALL of these people in plotter.who.''',
    'all_kinds': '''UNSET or str.
        include only plotters with ALL of these kinds in plotter.kinds.''',
    'skip_who': '''list
        exclude plotters with any of these people in plotter.who.''',
    'skip_kinds': '''list
        exclude plotters with any of these kinds in plotter.kinds.''',
    'min_cost': '''None or number
        exclude plotters with cost < min_cost.
        None --> no minimum.''',
    'max_cost': '''None or number
        exclude plotters with cost > max_cost.
        None --> no maximum.''',
}


class PlotterManager(metaclass=MetaPlotterManager):
    '''class to help with managing plotters.'''

    def get_plotter(self, name, who=UNSET):
        '''gets the Plotter associated with this name and who.
        Roughly equivalent: self.KNOWN_PLOTTERS[(name, who)]

        name: str or Plotter
            name of the plotter to use, or a Plotter instance.
        who: UNSET, None, or str
            person associated with the plotter.
            UNSET --> use the plotter with this name; crash if found multiple same-named plotters.

        see also: self.get_plotters(), which is good if you don't know 'who', or want multiple plotters.
        '''
        if isinstance(name, Plotter):
            return name
        elif who is UNSET:
            plotters = self.get_plotters(name=name, who=who)
            if len(plotters) == 0:
                raise PlottingNotImplementedError(f'no KNOWN_PLOTTERS with name={name!r}.')
            elif len(plotters) >= 2:
                errmsg = f'multiple KNOWN_PLOTTERS with name={name!r}, and did not specify who. Found: {plotters}'
                raise PlottingAmbiguityError(errmsg)
            else:  # len(plotters) == 1
                return plotters[0]
        else:
            try:
                return self.KNOWN_PLOTTERS[(name, who)]
            except KeyError:
                errmsg = f'plotter (name={name!r}, who={who!r}) not found in KNOWN_PLOTTERS.'
                same_name_different_person = [k for k in self.KNOWN_PLOTTERS.keys() if k[0]==name]
                if len(same_name_different_person) > 0:
                    errmsg = errmsg + f'\n(maybe these plotters are relevant, though?: {same_name_different_person!r})'
                raise PlottingNotImplementedError(errmsg) from None

    @format_docstring(**_paramdocs_get_plotters, sub_ntab=1)
    def get_plotters(self, who=UNSET, kind=UNSET, *, name=None,
                     all_whos=UNSET, all_kinds=UNSET, skip_who=[], skip_kinds=[],
                     min_cost=None, max_cost=None, sortby=None,
                     returns='plotters'):
        '''return list of all plotters associated with these inputs.
        If called with no inputs, just returns a copy of self.UNIQUE_PLOTTERS.

        who: {who}
        kind: {kind}
        name: {name}
        all_whos: {all_whos}
        all_kinds: {all_kinds}
        skip_who: {skip_who}
        skip_kinds: {skip_kinds}
        min_cost: {min_cost}
        max_cost: {max_cost}

        sortby: None or 'cost'
            tells how to sort the result (if returns=='plotters').
            None --> keep in order they appear in self.UNIQUE_PLOTTERS.
            'cost' --> sort by plotter.cost values.

        returns: 'plotters', 'names', 'who', 'kind', or 'kinds'
            'plotters' --> returns list of plotters
            'names' --> returns set of all names associated with at least 1 plotter in result.
            'who' --> returns set of all people associated with at least 1 plotter in result.
            'kind' or 'kinds' --> returns set of all kinds associated with at least 1 plotter in result.
        '''
        result = list(self.UNIQUE_PLOTTERS)
        if name is not None:
            result = [p for p in result if p.name==name]
        if min_cost is not None:
            result = [p for p in result if p.cost >= min_cost]
        if max_cost is not None:
            result = [p for p in result if p.cost <= max_cost]
        if who is not UNSET:
            if who is None or isinstance(who, str):
                who = [who]
            new_result = []
            for p in result:
                if len(p.who)==0 and None in who:
                    new_result.append(p)
                elif any(w in p.who for w in who):
                    new_result.append(p)
            result = new_result
        if kind is not UNSET:
            if isinstance(kind, str):
                result = [p for p in result if kind in p.kinds]
            else:
                result = [p for p in result if any(k in p.kinds for k in kind)]
        if all_whos is not UNSET:
            result = [p for p in result if all(w in p.who for w in all_whos)]
        if all_kinds is not UNSET:
            result = [p for p in result if all(k in p.kinds for k in all_kinds)]
        if skip_who:
            result = [p for p in result if not any(w in p.who for w in skip_who)]
        if skip_kinds:
            result = [p for p in result if not any(k in p.kinds for k in skip_kinds)]
        # returning result
        if returns == 'names':
            result = set(p.name for p in result)
        elif returns == 'who':
            result = set(w for p in result for w in p.who)
        elif returns == 'kind' or returns == 'kinds':
            result = set(k for p in result for k in p.kinds)
        elif returns == 'plotters':
            if sortby == 'cost':
                result = sorted(result, key=lambda p: p.cost)
        else:
            raise InputError(f'returns={returns!r}. Expected "plotters", "names", "who", "kind", or "kinds".')
        return result
    
    @format_docstring(**_paramdocs_plot, sub_ntab=1)
    def plot(self, name, who=UNSET, *, save=False, show=False, close=False, **kw_plotter):
        '''makes a single plot using the relevant plotter.

        name: str or Plotter
            name of the plotter to use, or a Plotter instance.
            (if Plotter, use directly and ignore 'who' input.)
        who: UNSET, None, or str
            person associated with the plotter.
            UNSET --> use the plotter with this name; crash if found multiple same-named plotters.
        save: {save}
        show: {show}
        close: {show}
        additional kwargs go to plotter.plot(...)

        see also: self.get_plotters(), self.save_plots().
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        plotter = self.get_plotter(name)
        kw = dict(save=save, show=show, close=close, **kw_plotter)
        return plotter.plot(self, **kw)

    @format_docstring(**_paramdocs_get_plotters, **_paramdocs_plot, sub_ntab=2)
    def save_plots(self, kind=UNSET, who=UNSET, *, name=None, 
                   all_whos=UNSET, all_kinds=UNSET, skip_who=[], skip_kinds=[],
                   min_cost=None, max_cost=None,
                   dst='{savename}', save_log=True, log_extras=[],
                   kw_save=dict(), bbox_inches=UNSET, dpi=UNSET,
                   show=False, close=True,
                   print_freq=0, **kw_plotter):
        '''saves all plots from plotters associated with these inputs.
        returns dict of {{plotter: plotter result}} for all plotters called.
        Consider checking self.get_plotters() first to learn which plotters will be included.

        For choosing which plotters to include:
            kind: {kind}
            who: {who}
            name: {name}
            all_whos: {all_whos}
            all_kinds: {all_kinds}
            skip_who: {skip_who}
            skip_kinds: {skip_kinds}
            min_cost: {min_cost}
            max_cost: {max_cost}

        For plotting:
            dst: str
                where to save plots to. Hit by dst.format(name=plotter.name, savename=plotter.savename).
                if not abspath, save to os.path.join(self.unique_notes_dirname, dst) if possible,
                    else save to dst within current directory.
            save_log: bool or str
                whether to save a log of plot progress to _save_plots_log.txt file.
                str --> save to this file name. If not abspath, put it in dir implied by dst (see above).
                The log tells current datetime, version info about PlasmaCalcs, and plot timing updates.
            log_extras: list of str
                extra lines to put in the log "header", if doing save_log.
            kw_save: dict
                kwargs to pass to plotter.save(...)
            bbox_inches: UNSET or any value
                if provided, added to kw_save.
            dpi: UNSET or any value
                if provided, added to kw_save.
            show: {show}
            close: {show}

            additional kwargs go to plotter.plot(...)

        Misc:
            print_freq: number
                minimum seconds between printing progress updates.
                -1 --> never print; 0 --> always print.

        see also: self.get_plotters(), self.plot().
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        result = {}
        plotters = self.get_plotters(name=name, who=who, kind=kind,
                                     all_whos=all_whos, all_kinds=all_kinds,
                                     skip_who=skip_who, skip_kinds=skip_kinds,
                                     min_cost=min_cost, max_cost=max_cost, sortby='cost')
        kw_save = kw_save.copy()
        if bbox_inches is not UNSET: kw_save.setdefault('bbox_inches', bbox_inches)
        if dpi is not UNSET: kw_save.setdefault('dpi', dpi)
        if not os.path.isabs(dst):
            dst = os.path.join(getattr(self, 'unique_notes_dirname', os.getcwd()), dst)
        if save_log == True:
            save_log = '_save_plots_log.txt'
        if save_log and not os.path.isabs(save_log):
            save_log = os.path.join(os.path.dirname(dst), save_log)
        save = dict(dst=dst, **kw_save)
        updater = ProgressUpdater(print_freq=print_freq)
        with contextlib.ExitStack() as context_stack:
            logfile = None
            try:
                for i, plotter in enumerate(plotters):
                    # logging progress
                    message = f'plotting {i+1:2d} of {len(plotters):2d}: {plotter.savename} (cost={plotter.cost})'
                    message = f'{message:<60} | '
                    updater.print(message)
                    full_message = updater.message_to_print(message)  # includes info about time elapsed.
                    # >> actually plotting <<
                    plotted = self.plot(plotter, save=save, show=show, close=close, **kw_plotter)
                    result[plotter] = plotted
                    # logging progress (only start logging after at least 1 plot completes successfully.)
                    if save_log:  # [TODO] encapsulate some of this to put it elsewhere...
                        if i==0:  # first time! open file and make "header"
                            logfile = context_stack.enter_context(open(save_log, 'a'))
                            codeinfo = code_snapshot_info()
                            lines = ['----------',
                                    codeinfo.pop("datetime"),
                                    'Created by PlasmaCalcs save_plots().',
                                    str(codeinfo),
                                    *log_extras]
                            logfile.write('\n'.join(lines) + '\n')
                        logfile.write(full_message + '\n')
            except BaseException as err:  # BaseException is okay because we raise immediately after logging.
                if logfile is not None:
                    logfile.write(f'CRASHED DUE TO ERROR: {err!r}\n')
                raise err
            finally:
                # logging progress
                if save_log:
                    updater.finalize('save_plots', always=True, file=logfile)
                    logfile.write('\n----------\n')
                updater.finalize('save_plots')
        return result



import math
import os

import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.backends.backend_pdf import PdfPages

import PlasmaCalcs

# SET PATH
path_to_folder = r"G:\My Drive\Research\eppic_out\eregion"
folder_names = {"FBI, normal": "FBI_normal", "FBI, Maxwell": "FBI_maxwell", "laminar, normal": "laminar_normal", "laminar, Maxwell": "laminar_maxwell"}  # key: nickname, value: folder name

# SET SPECIES
species_names = ("Electron", "Ion")  # list ALL species names IN THE SAME ORDER as in the simulation

# SET PLOT SPECIFICATIONS
moment_to_plot = (1,)  # 1: mean, 2: variance, etc.
plot_absolute_value = True
directions_to_plot = ("y",)

# SET PLOT CONFIGURATIONS
separate_species = True  # True: one species per plot, False: all species in one plot
separate_runs = False  # True: one run per plot, False: all runs in one plot

# SPECIAL CONFIGURATIONS
n_columns = 2  # number of plot columns
convert_variance_to_temperature = True  # plot variance in units of temperature instead of velocity squared

# SET PDF NAME
pdf_name = None  # Set to None for auto naming
if pdf_name is None:
    pdf_name = "_vs_".join(folder_names.keys())
    pdf_name += "_moment"
    for number in moment_to_plot:
        pdf_name += str(number)

# SET Y LABEL
y_label = None  # Set to None for auto naming
if y_label is None:
    if len(moment_to_plot) == 1:
        moment, = moment_to_plot
        if moment == 1:
            if plot_absolute_value:
                y_label = "|Velocity| (m/s)"
            else:
                y_label = "Velocity (m/s)"
        elif moment == 2:
            if convert_variance_to_temperature:
                y_label = "Temperature (K)"
            else:
                y_label = "Variance (m2 s-2)"
        elif moment == 3:
            y_label = "Third moment (m3 s-3)"
        elif moment == 4:
            y_label = "Fourth moment (m4 s-4)"
        else:
            assert False, "Invalid moment"
    else:
        y_label = "Value"


# FUNCTION TO GET PLOT TITLES
def _get_plot_titles(all_runs, all_species, number_of_columns, title_prefix="", title_suffix=""):
    """
    Get plot titles
    """
    n_sp = len(all_species)
    n_r = len(all_runs)
    title_list = []
    if separate_runs and n_r > 1:
        if separate_species and n_sp > 1:
            number_of_rows = math.ceil(n_sp * n_r / number_of_columns)
            for key in all_runs:
                for species in all_species:
                    title_list.append(f"{title_prefix} {species} ({key}) {title_suffix}")
        else:
            number_of_rows = math.ceil(n_r / number_of_columns)
            for key in all_runs:
                title_list.append(f"{title_prefix} {key} {title_suffix}")
    else:
        if separate_species and n_sp > 1:
            number_of_rows = math.ceil(n_sp / number_of_columns)
            for species in all_species:
                title_list.append(f"{title_prefix} {species} {title_suffix}")
        else:
            number_of_rows = 1
            number_of_columns = 1
            title_list.append(f"{title_prefix} {title_suffix}")

    # RESHAPE LIST
    reshaped_title_list = []
    title_counter = 0
    for _ in range(number_of_rows):
        temporary_list = []
        for _ in range(number_of_columns):
            if title_counter == len(title_list):
                temporary_list.append(None)
            else:
                temporary_list.append(title_list[title_counter])
                title_counter += 1
        reshaped_title_list.append(temporary_list)

    return reshaped_title_list


# SET PLOT TITLE
plot_titles = None  # Set to None for auto naming
title_prefix = "3D"
title_suffix = "drift parallel to E0"
if plot_titles is None:
    plot_titles = _get_plot_titles(folder_names, species_names, n_columns, title_prefix=title_prefix, title_suffix=title_suffix)
    print(plot_titles)

# GET SHAPE
n_rows = len(plot_titles)
n_columns = len(plot_titles[0])

# CREATE CANVAS
fig, axs = plt.subplots(n_rows, n_columns, figsize=[15, 8.5], squeeze=False)

# SET PDF PATH
pdf_path = os.path.join(path_to_folder, list(folder_names.values())[-1], f"{pdf_name}.pdf")

# SET COUNTERS
this_row = 0
this_column = 0

# NAME COLUMNS IN moments{fluid}.out (MIGHT NOT BE THE SAME FOR 2D RUNS!!!)
column_names = ['Index', 'Vx_Mean', 'Vx_Variance', 'Vx_Moment3', 'Vx_Moment4', 'Vy_Mean', 'Vy_Variance',
                'Vy_Moment3', 'Vy_Moment4', 'Vz_Mean', 'Vz_Variance', 'Vz_Moment3', 'Vz_Moment4']
assert len(column_names) == len(set(column_names)), "Duplicate column names"

# GET THE NAMES OF USED COLUMNS
columns_to_use = []
columns_to_convert_to_temperature = []
direction_dict = {"x": 0, "y": 4, "z": 8}
for moment in moment_to_plot:
    assert moment in (1, 2, 3, 4), "Invalid Moment"
    for direction in directions_to_plot:
        index = direction_dict[direction] + moment
        column_name = column_names[index]
        columns_to_use.append(column_name)
        if moment == 2 and convert_variance_to_temperature:
            columns_to_convert_to_temperature.append(column_name)

# MAKE THE PDF
with PdfPages(pdf_path) as pdf:

    # LOOP THROUGH EACH RUN
    for run in folder_names:

        # GET PATH
        this_path = os.path.join(path_to_folder, folder_names[run])

        # LOAD INPUT DECK
        ec = PlasmaCalcs.EppicCalculator.from_here(os.path.join(this_path,"eppic.i"), kw_units={"M": 1}, u_l=1)

        # GET NUMBER OF FLUIDS
        n_fluids = len(ec.fluids)

        # GET TIMES
        times = ec.snaps.t

        # GET kB
        kb = ec.u("kB")

        # LOOP THROUGH EACH SPECIES
        for fluid in range(n_fluids):
            # GET PATH TO FILE
            file_path = os.path.join(this_path, "domain000", f"moments{fluid}.out")

            # LOAD DATA
            data = pd.read_csv(file_path, sep='\s+', header=None, skiprows=1,
                               names=column_names, usecols=columns_to_use)

            # SPECIFY AXIS
            this_axis = axs[this_row, this_column]

            # SET TIMES
            data["Time"] = times

            # GET SPECIES MASS
            m = ec.fluids.m[fluid]

            # CONVERT VARIANCE TO TEMPERATURE
            factor = m / kb
            for column_name in columns_to_convert_to_temperature:
                data[column_name] *= factor

            # CONVERT TO ABSOLUTE VALUE
            if plot_absolute_value:
                for column_name in columns_to_use:
                    data[column_name] = abs(data[column_name])

            # RENAMING COLUMNS (FOR LEGENDS)
            new_names = {}
            if separate_species:
                if separate_runs:
                    for old_name in columns_to_use:
                        new_names[old_name] = old_name
                else:
                    for old_name in columns_to_use:
                        new_names[old_name] = f"{old_name} ({run})"
            else:
                if separate_runs:
                    for old_name in columns_to_use:
                        new_names[old_name] = f"{species_names[fluid]} {old_name}"
                else:
                    for old_name in columns_to_use:
                        new_names[old_name] = f"{species_names[fluid]} {old_name} ({run})"

            data.rename(columns=new_names, inplace=True)

            # PLOT DATA
            data.plot(x="Time", ax=this_axis)

            # SET TITLE AND LABELS
            this_axis.set_title(plot_titles[this_row][this_column])
            this_axis.set_xlabel("Time (s)")
            this_axis.set_ylabel(y_label)

            # ADVANCE AXIS
            if separate_species:
                if this_column == n_columns - 1:
                    this_row += 1
                    this_column = 0
                else:
                    this_column += 1

        # ADVANCE AXIS
        if separate_runs:
            if this_column == n_columns - 1:
                this_row += 1
                this_column = 0
            else:
                this_column += 1

        else:
            this_row = 0
            this_column = 0

    # MAKE LOWER LEFT (0, 0)
    for each_row in range(n_rows):
        for each_column in range(n_columns):
            this_axis = axs[each_row, each_column]
            this_axis.set_xlim(left=0)
            this_axis.set_ylim(bottom=0)

    # MAKE SURE PLOTS DON'T OVERLAP
    plt.tight_layout()

    # SAVE TO PDF
    pdf.savefig()
    plt.close()
    print(f"File saved to {pdf_path}.")



"""
File Purpose: code associated with Muram Snaps
"""
import os
import re

from .muram_io_tools import (
    muram_snap_files,
    read_muram_header,
    muram_directly_loadable_vars,
)
from ...dimensions import ParamsSnap, ParamsSnapList


''' --------------------- MuramSnap and MuramSnapList --------------------- '''

class MuramSnap(ParamsSnap):
    '''info about a single muram snapshot, including label (str) and index (int).

    The "index" should only be meaningful in the context of a SnapList
    The "label" should be the str name for this snapshot
        - unique within context (e.g. there's only one "snapshot 1" in a simulation)
        - easiest to use str int labels (e.g. "snapshot 1" --> label="1")

    s: the label (str) of the snapshot. if None, cannot convert to str.
    i: the "index" (int) of the snapshot (within a SnapList). if None, cannot convert to int.
    t: time at this snapshot ['raw' units]
    params: dict of parameters (e.g. from .idl file) at this snapshot. if None, make empty dict.
    '''
    @classmethod
    def from_header_file(cls, filename, *, i=None):
        '''return MuramSnap based on Header.NNN file.
        filename: str
            file to read from. Like Header.NNN
            will use s=NNN as label. E.g. "Header.004321" --> label="004321"
        i: None or int
            index of this snapshot in a SnapList.
        '''
        params = read_muram_header(filename)
        s = params['snap_s']
        t = params['t']
        return cls(s=s, i=i, t=t, params=params)

    @property
    def filename(self):
        '''return the basename of snapshot's header file: Header.NNN'''
        return f'Header.{self.s}'
    @property
    def dirname(self):
        '''return the directory containing this snapshot's Header.NNN file.'''
        return self.params['dirname']  # equivalent: self['dirname']
    @property
    def filepath(self):
        '''return the abspath to this snapshot's Header.NNN file.'''
        return os.path.join(self.dirname, self.filename)

    def directly_loadable_vars(self):
        '''return the list of directly loadable vars for this snap.'''
        return muram_directly_loadable_vars(self.s, dir=self.dirname)


class MuramSnapList(ParamsSnapList):
    '''list of MuramSnap objects'''
    value_type = MuramSnap

    @classmethod
    def from_header_files(cls, filenames):
        '''return MuramSnapList from a list of header files.
        order within result matches order of filenames.

        filenames: list of str
            list of filenames to read from. Like Header.NNN
            will use s=NNN as label. E.g. "Header.004321" --> label="004321"
        '''
        snaps = [cls.value_type.from_header_file(f, i=i) for i, f in enumerate(filenames)]
        return cls(snaps)

    @classmethod
    def from_here(cls, *, dir=os.curdir):
        '''return MuramSnapList from all Header.NNN files in directory.
        Sorted by snap number.
        dir: str. directory to look in.
        '''
        filenames = muram_snap_files(dir=dir)
        return cls.from_header_files(filenames)



"""
File Purpose: plotting lines vs time for xarray.
"""
import itertools
#import warnings

import cycler
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr

from .plot_settings import PlotSettings, PlotSettingsMixin
from .plot_tools import (
    current_axes_has_data,
    calculate_vlims, plt_zoomy,
)
from ..errors import (
    InputError, InputConflictError,
    PlottingAmbiguityError, TooManyPlottablesError,
    DimensionKeyError, DimensionError,
)
from ..tools import (
    UNSET,
    product, rargmax,
    pcAccessor, xarray_fill_coords, xarray_is_sorted, xarray_werr2pmstd,
    xarray_nondim_coords,
)

from ..defaults import DEFAULTS


class IndexableCycler(cycler.Cycler):
    '''Cycler which can be indexed by integers.
    Uses infinite indexing, i.e. result will be the same as the i'th value from itertools.cycle().
        [EFF] uses i % len(self), so the indexing time complexity doesn't scale with i.
    '''
    def __getitem__(self, i):
        '''return the ith element of the cycler, if i is an int.
        Otherwise use super().__getitem__.'''
        if isinstance(i, int):
            return list(self)[i % len(self)]
        else:
            return super().__getitem__(i)
        

''' --------------------- XarrayTimelines --------------------- '''

@pcAccessor.register('timelines')
@PlotSettings.format_docstring(
        default_cycles=[DEFAULTS.PLOT.TIMELINES_CYCLE0, DEFAULTS.PLOT.TIMELINES_CYCLE1],
        default_dmax=DEFAULTS.PLOT.TIMELINES_DMAX,)
class XarrayTimelines(PlotSettingsMixin):
    '''plotting lines vs time for an xarray.

    array: xarray.DataArray or xarray.Dataset
        the array to plot.
        Must have coord or dimension with name self.t (probably 't' or 'snap').
        if dataset, will be converted to DataArray with new dim named 'variable'.
    t: str
        name for the time coordinate (to be plotted along the x axis).
        if 'snap', use int(snap) for snap in array.coords['snap'].
        if 'snap_str', use int(str(snap)) for snap in array.coords['snap']
        if looks like '{{cname}}_index' or 'log_{{cname}}', and not in array yet,
            but cname in array, array.pc.fill_coords() to infer the t coord to use.
        if any other value, use array.coords[t].values.
        if array.coords[t].dtype==object, convert to strs when plotting.
            e.g. [Fluid('e'), Fluid('H+')] --> ticks with 'e', 'H+'.
    dims: None or list-like of str/None objects.
        the dimensions; plot one line for each combination of dimensions.
        None --> infer, based on array and t.
        list-like --> if any element is None, infer it based on the other dims.
    werr: bool, 'bar', or 'fill'
        if `array` is a Dataset with info about mean and std, use this to make plot with error bars.
            (else, werr=False is the only valid option.)
        True --> equivalent to 'bar'
        'bar' --> plt.errorbar. Expect `array` to have 'mean' and 'std' data_vars.
        'fill' --> plt.fill_between. Expect `array` to have one of the following pairs of data vars:
            ('mean' and 'std', 'mean+std' and 'mean-std', 'eval+std' and 'eval-std')
        Note: if using werr, can pass any kwargs here, (of plt.errorbar or plt.fill_between) for style.
        E.g. capsize=5, elinewidth=3, capthick=5
    fill_center: None, bool, or dict
        if True or provided, and werr='fill' and array has 'mean' (or 'eval') data_var, also make timeline at center,
        using kwargs from __init__ but overwriting some things for the internal call to timelines():
            anything provided in fill_center; werr=False, fill_center=False; alpha=1.
            (The alpha=1 prevents fill_center from appearing transparent if alpha<1 in __init__.)
    cycles: None or list-like of dict/Cycler/None objects.
        the styles to use for each dimension. Must have length >= len(dims).
        E.g. cycles=[dict(color=['r', 'b', 'g']), dict(linestyle=['-', '--'])]
            --> dims[0] will cycle through colors; dims[1] will cycle through linestyles.
        None --> use DEFAULTS.PLOT.TIMELINES_CYCLE0, and (if needed) DEFAULTS.PLOT.TIMELINES_CYCLE1.
                (default: {default_cycles})
                if len(dims) > 2, this will fail.
        list-like --> if any element is None, use plt.rcParams['axes.prop_cycle'] for that dimension;
                        and cannot have more than 1 None cycle.
    short_cycle_ok: bool
        whether it is okay for cycles to be shorter than the number of points in that dimension.
    dmax: UNSET, None, or int
        maximum length of a timelines dimension before crashing (with TooManyPlottablesError).
        This prevents accidentally asking to make plot with hundreds of lines or more,
            e.g. if array has maindims in it due to forgetting to use 'mean_'.
        Not applied to self.t dimension, which can be any length,
            since t goes along an axis instead of having 1 line for each.
        UNSET --> self.dmax = DEFAULTS.PLOT.TIMELINES_DMAX (default: {default_dmax})
        None --> no maximum. Use with caution.
        int --> maximum length.
    styles: None or list-like of dicts.
        the styles to use for each line. i'th line will use styles[i].
        if any value in styles dicts conflicts with cycles, use value from syles instead of cycle.
    plot: bool
        whether to call self.plot() immediately, during init.
    label: None, False, or str
        if provided, prepend this value to all lines' labels generated by this XarrayTimelines.
        (If it doesn't end with whitespace, add a single space. E.g. "myinfo" --> "myinfo ".)
        Useful if plotting timelines on the same axes as other information.
        False --> use label='' for all lines.
    custom_labels: None or str
        if provided, use this label instead of `label` + dim info;
        will be passed dict of nondim coord values for each line.
        E.g. custom_labels='u_{{fluid}},{{component}}' --> 'u_H+,x' when fluid='H+' and component='x'.
    add_legend: bool
        whether to plot a legend, by default.
    legend_cols_align: bool
        whether to align legend cols such that one of the dims (the longest one), is the same in each row.
        E.g. if True, and self.dims = ['fluid', 'component'], with more fluids than components,
            then 'fluid' will be the same across each row.
        Only applied if len(self) > legend_max_rows > (number of rows when cols aligned)
    legend_max_rows: None or int
        maximum number of rows in the legend. Add legend columns if len(self) > legend_max_rows.
        (For more precise control, make your own legend after plotting...)
        None --> no maximum.
    skipna: bool
        whether to drop NaN values before plotting each line of values.
    robust: {robust}
        Will consider minimum vmin and maximum vmax across all lines, to avoid fully-hiding lines.
    ymargin: {ymargin}
    ybounds: {ybounds}
    xincrease: None or bool
        whether the x-axis should increase from left to right.
        None --> False if monotonically nonincreasing. I.e., False if all(t[i+1] <= t[i]), else True.
                ("if t values are obviously inverted, take it as a hint from user to do xincrease=False")
    cstyles: None or dict of {{coordname: dict or list of tuples of (val, dict of kwargs for a line)}}
        if provided, pass these dicts to individual lines with corresponding scalar val for coord.
        use tuples of values to test equality instead of indexing a dict.
        E.g., styles={{'fluid': [('e', dict(ls='--')), ('H_II', dict(color='blue'))]}}
            would ensure dashed line when arr['fluid']=='e', blue line when arr['fluid']=='H_II',
            and have no effect whenever arr['fluid'] is not scalar, doesn't exist, or isn't 'e' or 'H_II'.
    cstyles_default: bool
        tells how to handle conflict between kwargs from cstyles and other kwargs.
        True --> treat cstyles as 'defaults'; kwargs from other sources override kwargs from cstyles.
        False --> kwargs from cstyles take precedence.

    Additional kwargs go to super().__init__, and eventually to a plotting routine. Options include:
        - self.plot_settings.get_mpl_kwargs('plt.legend')
        - self.plot_settings.get_mpl_kwargs('plt.plot')
        - self.plot_settings.get_mpl_kwargs('plt.errorbar')   # if werr=True or 'bar'
        - self.plot_settings.get_mpl_kwargs('plt.fill_between')   # if werr='fill'
        
        (Note that any plot settings which also appear in cycles will use the cycles values,
            e.g. if 'linestyle' appears in cycle but also in kwargs, use the cycle value instead.)

    --- Examples ---
        import PlasmaCalcs as pc

        # basic usage:
        array.pc.timelines()   # make timelines plot with default settings.

        # alternative call signature:
        pc.XarrayTimelines(array)   # same result as above ^

        # specify cycle:
        #   for dims[0], use plt.rcParams['axes.prop_cycle'];
        #   for dims[1], use a gradient in alpha (instead of the default, which is different linestyles):
        array.pc.timelines(cycles=[None, dict(alpha=np.linspace(1, 0.3, 10)])

        # use linewidth and linestyle in cycle for dims[1], use colors for dims[0]:
        array.pc.timelines(cycles=[None, dict(linewidth=[5, 4, 3, 2], linestyle=['-', '--', ':', '-.'])])

        # specify handlelength (for legend, ensure lines are long enough to see):
        array.pc.timelines(handlelength=5)
    '''
    def __init__(self, array, t='t', dims=None, *,
                 werr=False, fill_center=None,
                 cycles=None, short_cycle_ok=False, dmax=UNSET,
                 styles=None,
                 plot=True,
                 label=None,
                 custom_labels=None,
                 add_legend=True, legend_cols_align=True, legend_max_rows=20,
                 skipna=False,
                 robust=PlotSettings.default('robust'),
                 ymargin=PlotSettings.default('ymargin'),
                 ybounds=PlotSettings.default('ybounds'),
                 xincrease=None,
                 cstyles=None, cstyles_default=False,
                 **kw_super):
        self._kw_init = dict(t=t, dims=dims, werr=werr, fill_center=fill_center,
                             cycles=cycles, short_cycle_ok=short_cycle_ok, dmax=dmax,
                             styles=styles, plot=plot, label=label,
                             add_legend=add_legend, legend_cols_align=legend_cols_align,
                             legend_max_rows=legend_max_rows, skipna=skipna,
                             robust=robust, ymargin=ymargin, ybounds=ybounds,
                             xincrease=xincrease, cstyles=cstyles, cstyles_default=cstyles_default,
                             **kw_super)
        self.werr = werr
        self.fill_center = None if fill_center is False else dict() if fill_center is True else fill_center
        array = xarray_fill_coords(array, need=[t])
        if werr:
            self.werrarr = array
            if not isinstance(array, xr.Dataset):
                raise InputConflictError('werr=True is only valid for Dataset inputs.')
            if werr == True or werr == 'bar':
                if 'mean' not in array or 'std' not in array:
                    raise InputConflictError('werr=True is only valid for Dataset with "mean" and "std" data_vars.')
                self.array = array['mean']
                self.werr_std = array['std']
            elif werr == 'fill':
                # fillarr = array but formatted to have 'mean+std' and 'mean-std' values.
                if 'mean' in array and 'std' in array:
                    werrarr = xarray_werr2pmstd(array, keep_mean=True)
                elif 'mean+std' in array and 'mean-std' in array:
                    werrarr = array
                elif 'eval+std' in array and 'eval-std' in array:
                    werrarr = array.rename({'eval+std': 'mean+std', 'eval-std': 'mean-std'})
                    if 'eval' in array:
                        werrarr = werrarr.rename({'eval': 'mean'})
                else:
                    errmsg = ('werr="fill" is only valid for Dataset with one of these pairs of data_vars: '
                                '("mean" and "std", "mean+std" and "mean-std", or "eval+std" and "eval-std"), '
                                f'but got Dataset with data_vars: {list(array.data_vars)}')
                    raise InputConflictError(errmsg)
                if not (xarray_is_sorted(werrarr[t]) or xarray_is_sorted(werrarr[t], increasing=False)):
                    #warnings.warn(f'werr="fill" but array[{t!r}] not sorted! plt.fill_between might not work nicely.')
                    werrarr = werrarr.sortby(t)
                self.werrarr = werrarr
                self.array = werrarr['mean+std']  # used internally for coord & dim info
        else:
            if isinstance(array, xr.Dataset):
                array = array.to_array(dim='variable')
            self.array = xarray_fill_coords(array, need=[t])
        self.t = t
        self.dims = self._infer_dims(dims)
        self.short_cycle_ok = short_cycle_ok
        self.cycles = cycles   # after short_cycle_ok in case super() defines cycles as a property.
        self.dmax = DEFAULTS.PLOT.TIMELINES_DMAX if dmax is UNSET else dmax
        self.styles = styles
        self.label = label
        self.custom_labels = custom_labels
        self.add_legend = add_legend
        self.legend_cols_align = legend_cols_align
        self.legend_max_rows = legend_max_rows
        self.skipna = skipna
        self.cstyles = cstyles
        self.cstyles_default = cstyles_default
        super().__init__(robust=robust, ymargin=ymargin, ybounds=ybounds,
                         xincrease=xincrease, **kw_super)
        if plot:
            self.plot()

    def _infer_dims(self, dims):
        '''infer dims from self.array, as needed. return result.'''
        coords = self.array.coords
        if self.t == 'snap_str':
            raise NotImplementedError('[TODO] implement t == "snap_str" case.')
        tdim = coords[self.t].dims[0]
        rdims = [d for d in coords.dims if d != tdim]  # remaining dims (here, without tdim)
        if dims is None:
            return rdims
        else:
            dims = list(dims)  # editable copy of input dims.
            # first, ensure any non-None dims are in rdims, then remove them from rdims.
            for d in dims:
                if d is not None:
                    if d not in rdims:
                        raise DimensionKeyError(f'provided dim {d!r} not in remaining dims {rdims!r}')
                    else:
                        rdims.remove(d)
            # next, replace any None dims with values from rdims;
            #   and assert len(rdims) == number of None dims remaining.
            num_None = dims.count(None)
            if num_None != len(rdims):
                raise PlottingAmbiguityError(f'cannot infer dims {dims!r} from array with dims {coords.dims!r}')
            for i, d in enumerate(dims):
                if d is None:
                    dims[i] = rdims.pop(0)  # pop the 0th element from rdims.
            return dims

    def _get_cyclers(self, called=False, *, indexable=True):
        '''return list of Cycler objects. Get from self.cycles, or defaults if self.cycles is None.

        result must have length >= len(self.dims).

        Each cycle must have length >= number of points in the corresponding dimension,
            unless self.short_cycle_ok is True.

        called: bool
            if True, call each cycle before returning it. (Turning it into an itertools.cycle() result)
            Equivalent to [cycle() for cycle in self._get_cyclers(called=False)].
        indexable: bool
            if True, convert each result to IndexableCycler. Ignored if called=True.
        '''
        if self.cycles is None:
            cycles = [DEFAULTS.PLOT.TIMELINES_CYCLE0, DEFAULTS.PLOT.TIMELINES_CYCLE1]
        else:
            cycles = self.cycles
        cycles = list(cycles)   # make an editable copy
        if len(cycles) < len(self.dims):
            errmsg = f'len(cycles) < len(dims). ({len(cycles)} < {len(self.dims)})'
            raise PlottingAmbiguityError(errmsg)
        used_rcParams_cycle = False
        result = []
        for i, (cycle, dim) in enumerate(zip(cycles, self.dims)):
            if isinstance(cycle, cycler.Cycler):
                c_use = cycle
            else:
                if cycle is None:
                    if used_rcParams_cycle:
                        errmsg = 'cannot have more than 1 None cycle, otherwise plot will be ambiguous.'
                        raise PlottingAmbiguityError(errmsg)
                    c_use = plt.rcParams['axes.prop_cycle']
                else:
                    c_use = cycler.cycler(**cycle)
            assert isinstance(c_use, cycler.Cycler), "if False, probably made a coding error..."
            result.append(c_use)
            if (not self.short_cycle_ok) and (len(c_use) < len(self.array.coords[dim])):
                errmsg = (f'len(cycles[{i}]) < len(array.coords[{dim!r}]). ({len(result[i])} < {len(self.array.coords[dim])})'
                          f'\nUse short_cycle_ok=True to allow this, causing repeated line formats, instead of crashing.')
                raise PlottingAmbiguityError(errmsg)
        if called:
            result = [cycle() for cycle in result]
        elif indexable:
            result = [IndexableCycler(cycle) for cycle in result]
        return result

    def _get_t_values(self):
        '''return values for the time axis.'''
        if self.t == 'snap':
            return [int(snap) for snap in self.array.coords['snap']]
        elif self.t == 'snap_str':
            return [int(str(snap)) for snap in self.array.coords['snap']]
        else:
            cc = self.array.coords[self.t]
            if cc.dtype == object:
                result = [str(val) for val in cc.values]
                MAXLEN = DEFAULTS.PLOT.XTICKLABEL_MAXLEN
                result = [s if len(s)<MAXLEN else s[:MAXLEN-3]+'...' for s in result]
                return result
            else:
                return cc.values

    def _check_size(self):
        '''crash (with TooManyPlottablesError), if the number of lines to plot is too large.
        I.e., if any of the dims have length > self.dmax.
        '''
        for dim in self.dims:
            if len(self.array.coords[dim]) > self.dmax:
                errmsg = (f'len(array.coords[{dim!r}]) > self.dmax. '
                          f'(Got {len(self.array.coords[dim])} > {self.dmax})')
                raise TooManyPlottablesError(errmsg)

    def _get_label(self, dimsel, *, fmtn='{:.3g}', fmts='{!s:s}'):
        '''return labels given this dict of {dim: (i for this dim)} for dims in self.dims.
        This corresponds to 1 timeline in the plot.

        if custom_labels provided instead format it, by providing the dict of
            self.array.sel(dimsel).pc.nondims_coords(scalars_only=True, item=True).

        fmtn: str
            attempt to format values using this format string.
            if fails (with ValueError or TypeError), use fmts instead.
        fmts: str
            format string to use if fmtn fails.
            (default '{!s:s}'... '!s' converts to str; ':s' is the format specifier)
        '''
        if self.custom_labels is None:
            # prepend
            prepend0 = '' if self.label is None else self.label  # prepend without whitespace appended.
            if prepend0 is False:
                return ''  # don't provide any label at all.
            prepend = prepend0  # prepend, possibly with whitespace appended.
            if len(prepend) > 0:
                if prepend.rstrip() == prepend:   # i.e., doesn't end with any whitespace
                    prepend += ' '  # add a single space.
            # loop through dims
            coords = self.array.coords
            coords_dict = {}
            for dim, i in dimsel.items():
                coords_dict[dim] = coords[dim][i].item()  # item --> get single value
            dimlabels = []
            for dim, val in coords_dict.items():
                try:
                    valstr = fmtn.format(val)
                except (ValueError, TypeError):
                    valstr = fmts.format(val)
                dimlabels.append(f'{dim}={valstr}')
            if len(dimlabels) == 0:
                return prepend0
            else:
                return prepend + ', '.join(dimlabels)
        else:  # provided custom labels
            arr_here = self.array.isel(dimsel)
            nondim_coords = xarray_nondim_coords(arr_here, scalars_only=True, item=True)
            return self.custom_labels.format(**nondim_coords)

    def __len__(self):
        '''returns number lines which self would plot if plotted.'''
        dimlens = [len(self.array.coords[dim]) for dim in self.dims]
        return product(dimlens)

    def legend(self, ncols=None):
        '''plot the legend for the current plot, via plt.legend().
        legend anchored to upper right corner of plot, outside the axes.
            if self.legend_max_rows is not None:
                put a limit on nrows.
            if self.legend_cols_align is True:
                align the columns such that (the longest) one of the dims is the same in each row.
                Only applied if len(self) > legend_max_rows > (number of rows when cols aligned)
        ncols: None or int
            number of columns to use in the legend.
            None --> picks a decent-looking value, based on self.legend_max_rows.

        For more precise legend control, use plt.legend() instead.
        '''
        handles, labels = plt.gca().get_legend_handles_labels()
        if ncols is None:
            if self.legend_max_rows is None:
                ncols = 1
            else:
                L = len(self)
                nrow = self.legend_max_rows
                # align cols if necessary & possible
                if self.legend_cols_align and (L > nrow):
                    L_coords = [len(self.array.coords[dim]) for dim in self.dims]
                    i_longest = rargmax(L_coords)
                    L_longest = L_coords[i_longest]
                    if L_longest <= self.legend_max_rows:
                        nrow = L_longest
                        # also, re-sort handles and labels if needed:
                        if i_longest != len(self.dims) - 1:
                            labarr = np.array([handles, labels], dtype=object)
                            labarr = labarr.reshape(2, *L_coords)
                            transp = list(range(1, len(self.dims)+1))  # shape[0] will always be 2 for [handles,labels]
                            transp.remove(i_longest+1)
                            labarr = labarr.transpose(0, *transp, i_longest+1)  # put longest last.
                            labarr = labarr.reshape(2, -1)
                            handles, labels = labarr
                # 1 + ((x-1) // y) is equivalent to ceil(x/y), for positive x,y.
                ncols = 1 + ((len(self)-1) // nrow)
        # if ncols > 1, don't use any vertical padding from top right corner.
        bbox_to_anchor = (1.02, 0.95) if ncols == 1 else (1.02, 1)
        kw_legend_here = dict(loc='upper left', bbox_to_anchor=bbox_to_anchor, ncols=ncols)
        kw_legend = self.plot_settings.get_mpl_kwargs('plt.legend', **kw_legend_here)
        plt.legend(handles, labels, **kw_legend)

    def add_labels(self):
        '''add labels to plot, based on self.array.'''
        array = self.array
        units = array.attrs.get('units', None)
        units = '' if units is None else f' [{units}]'
        xunits = units if units and (self.t not in ('snap', 'snap_str')) else ''
        plt.xlabel(f'{self.t}{xunits}')
        name = getattr(self.array, 'name', None)
        if name is not None:
            plt.ylabel(f'{name}{units}')

    def _get_ith_style(self, i):
        '''returns style for ith line on the plot, inferred from self.styles.'''
        if self.styles is None:
            return dict()
        elif i < len(self.styles):
            return self.styles[i]
        else:
            return dict()

    def _get_idx_cstyle(self, idx):
        '''returns the cstyle for plot line at this dimpoint index (a dict).'''
        if self.cstyles is None:
            return dict()
        arr = self.array.isel(idx)
        result = {}
        for cname, csty in self.cstyles.items():
            if cname in arr.coords and arr[cname].size==1:
                if isinstance(csty, dict):
                    result.update(csty.get(arr[cname].item(), {}))
                else:  # csty must be an iterable of 2-tuples
                    for cval, cvalstyle in csty:
                        if arr[cname].item() == cval:
                            result.update(cvalstyle)
        return result

    def iter_dimpoints(self):
        '''iterates through points across all dimensions.
        Yields (dimsel, style) pairs, where:
            dimsel = dict of {dim: (i of dim at this point)} for dim in self.dims
            style = dict of {matplotlib kwarg: value} which are relevant at this point.
                    only includes relevant cstyles if not self.cstyles_default.
        '''
        dimranges = [range(len(self.array.coords[dim])) for dim in self.dims]
        cycles = self._get_cyclers(indexable=True)
        for ii, dimidx in enumerate(itertools.product(*dimranges)):
            # dimidx is a tuple of (i of dim here) for dim in self.dims.
            dimpoint = {dim: i for dim, i in zip(self.dims, dimidx)}
            # styles to use here
            styles = [cycle[j] for cycle, j in zip(cycles, dimidx)]
            style = {k: v for d in styles for k, v in d.items()}
            style.update(self._get_ith_style(ii))
            if not self.cstyles_default:
                cstyle_here = self._get_idx_cstyle(dimpoint)
                style = {**style, **cstyle_here}
            yield (dimpoint, style)

    @PlotSettings.format_docstring(ntab=2)
    def get_effective_data_interval(self, *, robust=UNSET):
        '''returns (vmin, vmax) taken across all lines in self.
        If robust, take min vmin(line) and max vmax(line) across all lines.

        robust: {robust}
            If UNSET, use self.plot_settings.get('robust'). (if still UNSET, use behavior described above.)
        '''
        self._check_size()
        array = self.werrarr if self.werr else self.array
        robust = self.plot_settings.get('robust', robust)
        yminmin, ymaxmax = None, None
        for dimsel, _style in self.iter_dimpoints():
            consider = array.isel(dimsel)
            ndim = consider.ndim if isinstance(consider, xr.DataArray) else len(consider.dims)
            if ndim != 1:
                raise DimensionError(f'expected 1D after indexing; got {consider.ndim}D')
            if self.werr:
                if self.werr == 'fill':
                    consider = xr.Dataset({'mean+std': consider['mean+std'], 'mean-std': consider['mean-std']})
                else:
                    consider = xarray_werr2pmstd(consider)
                consider = consider.to_array('__internal_dim_for_computing_data_interval__')
            vmin, vmax = calculate_vlims(consider, robust=robust)
            if yminmin is None:  # first time in loop.
                yminmin, ymaxmax = vmin, vmax
            else:
                yminmin = min(yminmin, vmin)
                ymaxmax = max(ymaxmax, vmax)
        return yminmin, ymaxmax

    @PlotSettings.format_docstring(ntab=2)
    def set_ylim(self, *, robust=UNSET, ymargin=None, ybounds=UNSET):
        '''sets plt.ylim() to a nice range for viewing all lines from self.
        robust: {robust}
            If robust, take min vmin(line) and max vmax(line) across all lines.
            If UNSET, use self.plot_settings.get('robust'). (if still UNSET, use behavior described above.)
        ymargin: {ymargin}
            If None, use self.plot_settings.get('ymargin') instead.
        ybounds: {ybounds}
            if UNSET, use self.plot_settings.get('ybounds') instead.
        '''
        data_interval = self.get_effective_data_interval(robust=robust)
        ymin, ymax = plt_zoomy(margin=ymargin, data_interval=data_interval, plot_settings=self.plot_settings)
        ybounds = self._get_ybounds(ybounds=ybounds)
        if ybounds is not False:
            ybmin, ybmax = ybounds
            if ybmin is not None: ymin = min(ymin, ybmin)
            if ybmax is not None: ymax = max(ymax, ybmax)
        plt.ylim(ymin, ymax)

    @PlotSettings.format_docstring(ntab=2)
    def _get_ybounds(self, *, ybounds=UNSET):
        '''get ybounds to use later; call before plotting, then pass to self.set_ylim after plotting
        to ensure any previous plots on this axes don't get cut off.
        robust: {robust}
            If robust, take min vmin(line) and max vmax(line) across all lines.
            If UNSET, use self.plot_settings.get('robust'). (if still UNSET, use behavior described above.)
        ymargin: {ymargin}
            If None, use self.plot_settings.get('ymargin') instead.
        ybounds: {ybounds}
            if UNSET, use self.plot_settings.get('ybounds') instead.
        '''
        if ybounds is UNSET:
            ybounds = self.plot_settings.get('ybounds')
        if ybounds is None:
            ybounds = plt.ylim() if current_axes_has_data() else False
        return ybounds

    def _plot_lines(self, **kw_plot):
        '''plot the lines vs time. Probably via plt.plot.
        (if self.werr, might instead use plt.errorbar or plt.fill_between.)

        Does not do any of the nice preprocessing or postprocessing to help with formatting.
        User should use self.plot() instead.

        returns dict of info, including 'any_legend_labels' telling whether any labels were added
            (if no labels were added, later functions should not add legend.)
        '''
        t_values_base = self._get_t_values()
        array = self.array
        any_legend_labels = False

        if self.werr == False:
            plotter_name = 'plt.plot'
        elif self.werr == True or self.werr == 'bar':
            plotter_name = 'plt.errorbar'
        elif self.werr == 'fill':
            plotter_name = 'plt.fill_between'
        else:
            raise InputError('invalid value for self.werr, expected bool, "bar" or "fill".')
        # plot lines
        for dimsel, style in self.iter_dimpoints():
            # get plottable values
            to_plot = array.isel(dimsel)
            if to_plot.ndim != 1:
                raise DimensionError(f'expected 1D array after indexing; got {to_plot.ndim}D array')
            if self.werr == True or self.werr == 'bar':
                yerr = self.werr_std.isel(dimsel)
            elif self.werr == 'fill':
                werrarr = self.werrarr.isel(dimsel)
            # account for nans
            t_values = t_values_base  # might be changed below
            if self.skipna:
                nans = np.isnan(to_plot)
                if np.any(nans):
                    to_plot = to_plot[~nans]
                    t_values = t_values_base[~nans]
                    if self.werr == True or self.werr == 'bar':
                        yerr = yerr[~nans]
                    elif self.werr == 'fill':
                        werrarr = werrarr.where(~nans, drop=True)
            # get label
            label = self._get_label(dimsel)
            if label != '':
                any_legend_labels = True
            # get kwargs to use for this line
            kw_plot_here = self.plot_settings.get_mpl_kwargs(plotter_name, **style, **kw_plot)
            if self.cstyles_default:  # put cstyle as defaults but allow plot_settings to override them
                cstyle_here = self._get_idx_cstyle(dimsel)
                kw_plot_here = {**cstyle_here, **kw_plot_here}
            # >> actually plot this line <<
            if self.werr == False:
                plt.plot(t_values, to_plot.values, label=label, **kw_plot_here)
            elif self.werr == True or self.werr == 'bar':
                plt.errorbar(t_values, to_plot.values, yerr=yerr.values, label=label, **kw_plot_here)
            else:
                lower, upper = werrarr['mean-std'].values, werrarr['mean+std'].values
                plt.fill_between(t_values, lower, upper, label=label, **kw_plot_here)
        return {'any_legend_labels': any_legend_labels}


    PlotSettings.format_docstring(ntab=2)
    def plot(self, *, legend=UNSET, add_labels=True, robust=UNSET, ymargin=None, ybounds=UNSET, **kw_plot):
        '''plot the lines vs time. Probably via plt.plot.
        (if self.werr, might instead use plt.errorbar or plt.fill_between.)

        legend: UNSET or bool
            whether to plt.legend().
            UNSET --> use legend = self.add_legend (default: True)
            note: if all labels are None or empty string, will never attempt to add legend.
        add_labels: bool
            whether to self.add_labels() to plot, based on self.array:
                xlabel (self.t), and ylabel if known (self.array.name).
                Also put units on those labels if self.array.attrs['units'] is provided.
        robust: {robust}
            If robust, take min vmin(line) and max vmax(line) across all lines.
            If UNSET, use self.plot_settings.get('robust'). (if still UNSET, use behavior described above.)
        ymargin: {ymargin}
            If None, use self.plot_settings.get('ymargin') instead.
        ybounds: {ybounds}
            if UNSET, use self.plot_settings.get('ybounds') instead.
        '''
        self._check_size()
        t_values_base = self._get_t_values()
        pre_ybounds = self._get_ybounds(ybounds=ybounds)
        # >> plot lines <<
        plotinfo = self._plot_lines()
        any_legend_labels = plotinfo.get('any_legend_labels', False)
        # formatting
        self.set_ylim(robust=robust, ymargin=ymargin, ybounds=pre_ybounds)
        xincrease = self.plot_settings.get('xincrease')
        if xincrease is None:
            if t_values_base[0] < t_values_base[-1]:  # [EFF] quick check with only 2 values
                xincrease = True
            else:  # False if all(t[i+1] <= t[i]), else True
                xincrease = not np.all(t_values_base[1:] <= t_values_base[:-1])
        xmin, xmax = plt.xlim()
        if (xincrease and (xmin > xmax)) or ((not xincrease) and (xmin < xmax)):
            plt.gca().invert_xaxis()
        if legend is UNSET: legend = self.add_legend
        if legend and any_legend_labels:
            self.legend()
        if add_labels:
            self.add_labels()
        # 'center' if fill_center
        if self.fill_center is not None:
            if self.werr != 'fill':
                raise InputConflictError(f'non-None fill_center when "fill"!=werr(=={self.werr}).')
            if 'mean' not in self.werrarr:
                raise InputConflictError('werrarr does not have "mean" data_var; cannot fill_center')
            timelines_cls = type(self)
            _default_center_kw = {'werr': False, 'fill_center': False, 'alpha': 1}
            center_kw = {**self._kw_init, **_default_center_kw, **self.fill_center}
            self.filled_center = timelines_cls(self.werrarr['mean'], **center_kw)



"""
File Purpose: loading info related to 'safe' eppic.i inputs,
from values in an EppicInstabilityCalculator.
"""
import numpy as np
import xarray as xr

from ..eppic_runtime_info import EppicRuntimeInfoLoader
from ..eppic_sim_info import EppicSimInfoLoader
from ....defaults import DEFAULTS
from ....errors import LoadingNotImplementedError
from ....tools import (
    simple_property,
    xarray_sum, xr1d,
)


class EppicSafetyInfoLoader(EppicSimInfoLoader, EppicRuntimeInfoLoader):
    '''loads info about 'safe' eppic.i inputs, from values in an EppicInstabilityCalculator.
    E.g. self('safe_sim_size') tells expected size of a simulation using 'safe' inputs.
        (including: safe_total_Gbytes, safe_pow2_Nspace, safe_spacetimesteps)
    '''

    # # # HACKY CLASS INHERITENCE STUFF # # #

    # required to ensure we don't use EppicSimInfoLoader's ndim_space.
    @known_var
    def get_ndim_space(self):
        '''number of spatial dimensions to put in the simulation. Loaded directly from self.ds.'''
        return self.load_direct('ndim_space')

    # required to ensure we don't use EppicSimInfoLoader's nptotcell
    @known_var(deps=['safe_nptotcelld', 'safe_pow2_subcycle'], aliases=['safe_nptotcell'])
    def get_nptotcell(self):
        '''number of particles per cell for all fluids combined, using 'safe' simulation inputs,
        appropriately weighted considering subcycling.
        Equivalent: sum_fluids_(safe_nptotcelld/safe_pow2_subcycle)
        '''
        return xarray_sum(self('safe_nptotcelld') / self('safe_pow2_subcycle'), 'fluid')

    # required to ensure we don't use EppicSimInfoLoader's nsubdomains
    @known_var(deps=['safe_nsubdomains'])
    def get_nsubdomains(self):
        '''number of subdomains to use for eppic run.
        Loaded directly from self.ds if available, else self('safe_nsubdomains')
        '''
        try:
            return self.load_direct('nsubdomains')
        except LoadingNotImplementedError:
            return self('safe_nsubdomains')


    # # # SIMULATION SIZE # # #

    SAFE_SIM_SIZE_VARS = [
        # size (time)
        'safe_dt', 'safe_nt', 'safe_rounded_nt',
        'safe_nout', 'safe_rounded_nout', 'safe_nsnaps',
        # size (space)
        'safe_dspace', 'safe_Nspace', 'safe_pow2_Nspace',
        'safe_data_array_Mbytes', 'arrays_per_snap',
        # size (total)
        'safe_total_Gbytes', 'safe_spacetimesteps',
        # size (runtime)
        'tasks_per_node',
        'n_nodes',
        'min_n_nodes', 'min_n_nodes_given_nsubdomains', 'min_n_nodes_given_runtime_guess',
        'safe_cpu_seconds_per_timestep', 'safe_cpu_seconds', 'safe_node_hours',
        'safe_runtime_seconds', 'safe_runtime_HMS',
    ]
    @known_var(deps=SAFE_SIM_SIZE_VARS)
    def get_safe_sim_size(self):
        '''dataset with various values related size of a simulation using 'safe' inputs.
        Equivalent: self(self.SAFE_SIM_SIZE_VARS)
        '''
        cached = dict(_cached_safe_dt=self('safe_dt'),
                      _cached_safe_pow2_subcycle=self('safe_pow2_subcycle'))
        with self.using(**cached):  # [EFF] caching on most expensive vars to avoid recalculating.
            return self(self.SAFE_SIM_SIZE_VARS)

    @known_var(deps=['safe_rounded_nt', 'safe_pow2_Nspace', 'ndim_space'])
    def get_safe_spacetimesteps(self):
        '''number of "cells" across space AND time for a simulation using 'safe' inputs.
        safe_spacetimesteps = safe_rounded_nt * (safe_pow2_Nspace ** ndim_space).
        '''
        timesteps = self('safe_rounded_nt') * 1.0   # *1.0 in case of Very Big results too large for int64.
        spacesteps = (self('safe_pow2_Nspace') * 1.0) ** self('ndim_space')
        return timesteps * spacesteps

    @known_var(deps=['safe_rounded_nt', 'safe_rounded_nout'])
    def get_safe_nsnaps(self):
        '''number of snapshots in entire simulation if using safe_rounded_nout and safe_nt.
        safe_nsnaps = safe_nt / safe_rounded_nout.
        '''
        return self('safe_rounded_nt') / self('safe_rounded_nout')

    @known_var(deps=['safe_pow2_Nspace', 'ndim_space', 'nout_avg'])
    def get_safe_data_array_size(self):
        '''number of elements in a single output data array for a simulation using 'safe' inputs.
        safe_data_array_size = (safe_pow2_Nspace/nout_avg) ** ndim_space
        '''
        return (self('safe_pow2_Nspace') / self('nout_avg')) ** self('ndim_space')

    @staticmethod
    def _to_bytes_factor(M=False, dtype='32'):
        '''number of (Mega)bytes per item in an array with this dtype.
        M: '', 'M', 'G'. Tells whether to use bytes, MB, or GB.
        dtype: '', '32', or '64'. Dtype to use. '' or '32' --> float32. '64' --> float64.
        '''
        if dtype=='' or dtype=='32':
            esize = np.dtype('float32').itemsize
        elif dtype=='64':
            esize = np.dtype('float64').itemsize
        if M=='M':
            esize = esize / 1024**2
        elif M=='G':
            esize = esize / 1024**3
        return esize

    @known_pattern(r'safe_data_array_(|M|G)bytes(|32|64)', deps=['safe_data_array_size'])
    def get_safe_data_array_bytes(self, var, *, _match=None):
        '''number of bytes in a single output data array for a simulation using 'safe' inputs.
        result = safe_data_array_size * number of bytes per element.
        pattern: Mbytes or Gbytes to get MB or GB. Uses 1024 not 1000 (e.g. MB --> 1024**2.)
        pattern: bytes32 or bytes64 to specify dtype. Unspecified --> float32.
        '''
        M, dtype = _match.groups()
        esize = self._to_bytes_factor(M, dtype)
        return self('safe_data_array_size') * esize

    @known_var(deps=['part_out_subcycle', 'flux_out_subcycle', 'nvsqr_out_subcycle'])
    def get_arrays_per_snap(self):
        '''average number of arrays (excluding vdist arrays) per snapshot file.
        n_global = 1   # phi, only.
        n_den = (nfluids / part_out_subcycle)
        n_flux = (nfluids * 3 / flux_out_subcycle)   # 3 is from vector components x,y,z
        n_nvsqr = (nfluids * 3 / nvsqr_out_subcycle)  # 3 is from vector components x,y,z
        n_snap_arrays_space = n_global + n_den + n_flux + n_nvsqr
        '''
        nfluids = len(self.fluids)
        n_global = 1
        n_den = nfluids / self('part_out_subcycle')
        n_flux = nfluids * 3 / self('flux_out_subcycle')
        n_nvsqr = nfluids * 3 / self('nvsqr_out_subcycle')
        return n_global + n_den + n_flux + n_nvsqr

    @known_pattern(r'safe_snap_(|M|G)bytes(|32|64)',
                   deps=['arrays_per_snap', {(0,1): 'safe_data_array_{group0}bytes{group1}'}])
    def get_safe_snap_bytes(self, var, *, _match=None):
        '''average number of bytes per snapshot (excluding vdist arrays) if using 'safe' inputs.
        safe_snap_bytes = arrays_per_snap * safe_data_array_bytes
        pattern: Mbytes or Gbytes to get MB or GB. Uses 1024 not 1000 (e.g. MB --> 1024**2.)
        pattern: bytes32 or bytes64 to specify dtype. Unspecified --> float32.
        '''
        M, dtype = _match.groups()
        n_per_snap = self('arrays_per_snap')
        arr_bytes = self(f'safe_data_array_{M}bytes{dtype}')
        return n_per_snap * arr_bytes

    @known_pattern(r'safe_total_(|M|G)bytes(|32|64)',
                   deps=['safe_nsnaps', {(0,1): 'safe_snap_{group0}bytes{group1}'}])
    def get_safe_total_bytes(self, var, *, _match=None):
        '''total number of bytes expected (excluding vdist) across all snaps, if using 'safe' inputs.
        safe_total_bytes = safe_nsnaps * safe_snap_bytes
        pattern: Mbytes or Gbytes to get MB or GB. Uses 1024 not 1000 (e.g. MB --> 1024**2.)
        pattern: bytes32 or bytes64 to specify dtype. Unspecified --> float32.
        '''
        M, dtype = _match.groups()
        nsnaps = self('safe_nsnaps')
        snap_bytes = self(f'safe_snap_{M}bytes{dtype}')
        return nsnaps * snap_bytes


    # # # NUMBER OF PROCESSORS TO USE # # #

    cls_behavior_attrs.register('tasks_per_node', default=56)
    tasks_per_node = simple_property('_tasks_per_node', default=56,
        doc='''number of processers per node to use for eppic run. Default 56 works well on Frontera.
        n_processors = n_nodes * n_tasks_per_node.
        Utilized by safe_runtime_seconds computations.''')

    @known_var
    def get_tasks_per_node(self):
        '''number of tasks per node to use for eppic run.
        internally, value is stored (and can be adjusted) at self.tasks_per_node.
        '''
        return xr.DataArray(self.tasks_per_node)

    cls_behavior_attrs.register('n_nodes', default=4)
    n_nodes = simple_property('_n_nodes', default=4,
        doc='''number of nodes to use for eppic run. Default 4.
        (n_processors = n_nodes * n_tasks_per_node.)
        Utilized by safe_cpu_seconds_... computations
            (because 'efield' and 'collect' runtime per node depends on n_nodes,
            because mpi communication is more expensive with more nodes).

        Can be set to DataArray if desired. Common examples:
            # minimum number of nodes 
            ec.n_nodes = ec('min_n_nodes')  # 
            ec.n_nodes = ec('min_n_nodes_given_nsubdomains')  # min allowed n_nodes (lowest node-hour cost)''')

    @known_var
    def get_n_nodes(self):
        '''number of nodes to use for eppic run.
        Larger is always more expensive (in cpu hours), but can reduce wall clock time
            (caution: large n_nodes can mean more nodes actually increases wall clock time,
                due to increased mpi communication costs.)
        internally, value is stored (and can be adjusted) at self.n_nodes.
            For more details see help(type(self).n_node) or self.help_call_options('n_nodes')
        '''
        return xr.DataArray(self.n_nodes)

    cls_behavior_attrs.register('max_runtime_hours', default=48)
    max_runtime_hours = simple_property('_max_runtime_hours', default=48,
        doc='''maximum allowed runtime, in wall clock hours. Default 48.
        Utilized by min_n_nodes computations.
            (less nodes always costs fewer node hours, but might take way too long).
        Can be set to DataArray if desired, to check multiple options at once.''')

    @known_var
    def get_max_runtime_hours(self):
        '''maximum allowed runtime, in wall clock hours.
        Utilized by min_n_nodes computations.
            (less nodes always costs fewer node hours, but might take way too long).
        internally, value is stored (and can be adjusted) at self.max_runtime_hours.
        '''
        return xr.DataArray(self.max_runtime_hours)

    cls_behavior_attrs.register('min_n_nodes_lims', default=(4, 2048))
    min_n_nodes_lims = simple_property('_min_n_nodes_lims', default=(4, 2048),
        doc='''two-tuple of min & max values to test for, during min_n_nodes computations.
        (4, 512) corresponds to (smallest, largest) power of 2 allowed in Frontera 'normal' queue.
            Default goes up to 2048 to allow for 'large' queue requests too.''')

    @known_var(deps=['safe_runtime_seconds', 'max_runtime_hours'])
    def get_min_n_nodes_given_runtime_guess(self):
        '''minimum number of nodes, given safe_runtime_seconds and max_runtime_hours,
        to ensure safe_runtime_seconds < max_runtime_hours * 3600,
            or use max_min_n_nodes if safe_runtime_seconds > max_runtime_hours.

        Only tests powers-of-2 between self.min_n_nodes_lims [0] and [1], inclusive.
        '''
        test_vals = [2**i for i in range(int(np.log2(self.min_n_nodes_lims[0])),
                                         int(np.log2(self.min_n_nodes_lims[1]) + 1))]
        test = xr1d(test_vals, '__checking__min_n_nodes_given_runtime_guess__')
        cache = dict(_cached_safe_pow2_subcycle=self('safe_pow2_subcycle'))  # [EFF] cache expensive vars
        #cache = dict()
        with self.using(n_nodes=test, **cache):
            safe_runtime_seconds = self('safe_runtime_seconds')
            max_runtime_seconds = self('max_runtime_hours') * 3600
        not_too_long = safe_runtime_seconds <= max_runtime_seconds
        result = test.where(not_too_long).min('__checking__min_n_nodes_given_runtime_guess__')
        if not result.isnull().any():   # store as 'int' if it won't override nan or inf values.
            result = result.astype('int')
        return result

    @known_var(deps=['min_n_nodes_given_runtime_guess', 'min_n_nodes_given_nsubdomains'])
    def get_min_n_nodes(self):
        '''minimum number of nodes, given runtime guess, max_runtime_hours, nsubdomains, and tasks_per_node.
        Ensures the following:
            (1) n_processors % nsubdomains == 0,
                where n_processors = n_nodes * tasks_per_node.
                (This is ensured by min_n_nodes_given_nsubdomains.)
            (2) safe_runtime_seconds < max_runtime_hours * 3600,
                or use max_min_n_nodes if safe_runtime_seconds > max_runtime_hours.
                (This is ensured by min_n_nodes_given_runtime_guess.)

        Equivalent: maximum(min_n_nodes_given_runtime_guess, min_n_nodes_given_nsubdomains)
        '''
        return np.maximum(self('min_n_nodes_given_runtime_guess'), self('min_n_nodes_given_nsubdomains'))


    # # # RUNTIME STUFF # # #

    cls_behavior_attrs.register('runtime_safety', default=DEFAULTS.EPPIC.RUNTIME_SAFETY)
    runtime_safety = simple_property('_runtime_safety', setdefault=lambda: DEFAULTS.EPPIC.RUNTIME_SAFETY,
        doc=f'''safety factor for safe runtime calculations. Larger is LESS safe.
        safe_cpu_seconds = unsafe_cpu_seconds * runtime_safety.
        Default: DEFAULTS.EPPIC.RUNTIME_SAFETY (default: {DEFAULTS.EPPIC.RUNTIME_SAFETY})''')

    @known_var
    def get_runtime_safety(self):
        '''safety factor for safe runtime calculations. Larger is LESS safe.
        safe_cpu_seconds = unsafe_cpu_seconds * runtime_safety.
        internally, value is stored (and can be adjusted) at self.runtime_safety.
        '''
        return self.runtime_safety

    @known_var(deps=['safe_pow2_Nspace', 'ndim_space', 'guess_cpu_seconds_sum_per_ct_safe', 'runtime_safety'])
    def get_safe_cpu_seconds_per_timestep(self):
        '''safety factors * cpu seconds expected per timestep, for simulation using 'safe' inputs.
        unsafe_result = (safe_pow2_Nspace ** ndim_space) * guess_cpu_seconds_sum_per_ct_safe
        safe_result = unsafe_result * runtime_safety.
        '''
        with self.using(_cached_safe_pow2_subcycle=self('safe_pow2_subcycle')):  # [EFF] cache avoids recalculating
            safe_ncells = self('safe_pow2_Nspace') ** self('ndim_space')
            unsafe_result = safe_ncells * self('guess_cpu_seconds_sum_per_ct_safe')
        return unsafe_result * self('runtime_safety')

    @known_var(deps=['safe_rounded_nt', 'safe_cpu_seconds_per_timestep'])
    def get_safe_cpu_seconds(self):
        '''safety factors * cpu seconds expected in total, for simulation using 'safe' inputs.
        result = safe_rounded_nt * safe_cpu_seconds_per_timestep

        To get required wall clock seconds, divide by number of processors.
        '''
        cached = dict(_cached_safe_pow2_Nspace=self('safe_pow2_Nspace'),
                      _cached_safe_dt=self('safe_dt'))
        with self.using(**cached):  # [EFF] caching on most expensive vars to avoid recalculating.
            return self('safe_rounded_nt') * self('safe_cpu_seconds_per_timestep')

    @known_var(deps=['n_processors', 'safe_cpu_seconds'])
    def get_safe_runtime_seconds(self):
        '''predicted run cost, in wall clock time [seconds]. safe_cpu_seconds / number of processors.
        n_processors = self.n_nodes * self.tasks_per_node.
        '''
        return self('safe_cpu_seconds') / self('n_processors')

    @known_var(deps=['safe_runtime_seconds'])
    def get_safe_runtime_HMS(self):
        '''predicted run cost, in wall clock time [string like: 'hh:mm:ss'].'''
        value = self('safe_runtime_seconds')
        hh_num = (value / 3600).astype('int')
        hh_str = hh_num.astype('str').str.zfill(2)
        mm_num = ((value - hh_num * 3600) / 60).astype('int')
        mm_str = mm_num.astype('str').str.zfill(2)
        ss_num = (value - hh_num * 3600 - mm_num * 60).astype('int')
        ss_str = ss_num.astype('str').str.zfill(2)
        colon_str = xr.DataArray(':')
        result = xr.concat([hh_str, colon_str, mm_str, colon_str, ss_str], dim='__to_join__', coords='minimal')
        return result.str.join('__to_join__')

    @known_var(deps=['safe_cpu_seconds', 'tasks_per_node'])
    def get_safe_node_hours(self):
        '''predicted run cost, in node hours. safe_cpu_seconds * seconds2hours / tasks_per_node.
        (seconds2hours = 1 / 3600)

        Compare directly with Frontera SU cost (1 SU is 1 node-hour in 'normal' partition).
        '''
        seconds2hours = 1 / 3600
        return self('safe_cpu_seconds') * seconds2hours / self('tasks_per_node')

    @known_var(deps=['safe_pow2_Nspace', 'guess_cpu_seconds_sum_per_ct'])
    def get_unsafe_cpu_seconds_per_timestep(self):
        '''cpu seconds expected per timestep, with no runtime_safety and no safety inside guess_cpu.
        safe_pow2_Nspace * guess_cpu_seconds_sum_per_ct  (without _safe)
        '''
        return self('safe_pow2_Nspace') * self('guess_cpu_seconds_sum_per_ct')

    @known_var(deps=['safe_cpu_seconds_per_timestep', 'unsafe_cpu_seconds_per_timestep'])
    def get_direct_runtime_safety(self):
        '''direct_runtime_safety = safe_cpu_seconds_per_timestep / unsafe_cpu_seconds_per_timestep.'''
        return self('safe_cpu_seconds_per_timestep') / self('unsafe_cpu_seconds_per_timestep')


    # # # SIMULATION SAFETY FACTORS DETAILS # # #
    # e.g. what were the input safety factors?
    # e.g. what were the effective, "direct", safety factors (accounting for pow2 rounding)?

    SAFETY_DETAILS_VARS = [
        'dspace_safety', 'direct_dspace_safety',
        'nspace_safety', 'direct_nspace_safety',
        'dt_safety', 'direct_dt_safety',
        'ntime_safety', 'direct_ntime_safety',
        'nout_waveprop_safety', 'direct_nout_waveprop_safety',
        'nout_growth_safety', 'direct_nout_growth_safety',
        # runtime
        'runtime_safety', 'direct_runtime_safety',
    ]
    @known_var(deps=SAFETY_DETAILS_VARS)
    def get_safety_details(self):
        '''dataset with various values related to safety factors used in 'safe' inputs.
        Equivalent: self(self.SAFETY_DETAILS_VARS)
        '''
        cached = dict(_cached_safe_pow2_Nspace=self('safe_pow2_Nspace'),
                        _cached_safe_dt=self('safe_dt'),
                        _cached_safe_pow2_subcycle=self('safe_pow2_subcycle'))
        with self.using(**cached):  # [EFF] caching on most expensive vars to avoid recalculating.
            return self(self.SAFETY_DETAILS_VARS)



"""
File Purpose: BifrostMultifluidCalculator
"""

from .bifrost_multifluid_densities import BifrostMultifluidDensityLoader
from ..bifrost_calculator import BifrostCalculator
from ....dimensions import SINGLE_FLUID
from ....mhd import MhdMultifluidCalculator, SpecieList
from ....tools import format_docstring

@format_docstring(bifrost_calculator_docs=BifrostCalculator.__doc__, sub_ntab=1)
class BifrostMultifluidCalculator(BifrostMultifluidDensityLoader,
                                  MhdMultifluidCalculator,
                                  BifrostCalculator):
    '''MultifluidPlasmaCalculator for Bifrost outputs.
    various possible ways to infer fluids.
        One possibility is to use abundances + saha ionization equation.
        [TODO] explain this in more detail.
        [TODO] allow to enter fluids as list of strings during init.

    set self.fluid=SINGLE_FLUID to get single-fluid values,
        otherwise will get inferred multifluid values.

    --- Docstring from BifrostCalculator copied below ---
        {bifrost_calculator_docs}
    '''

    SINGLE_FLUID = SINGLE_FLUID  # convenient reference to SINGLE_FLUID. Subclass should NEVER override.

    def __init__(self, snapname=None, *, units='si', **kw_super):
        super().__init__(snapname=snapname, units=units, **kw_super)
        self.init_fluids()

    # # # FLUIDS # # #
    def init_fluids(self):
        '''initialize self.fluids, fluid, jfluids, and jfluid.'''
        self.fluids = self.chromo_fluid_list()
        self.fluid = None
        self.jfluids = self.fluids
        self.jfluid = self.jfluids.get_neutral()

    def chromo_fluid_list(self):
        '''SpecieList of species relevant to the chromosphere, maybe.
        currently, just produces list of [electron, H_I, H_II, *other_once_ionized_ions]
            also includes He_III (after He_II) if self.params['do_helium'].
        '''
        elements = self.tabin.elements
        if (elements[0] != 'H') or (elements[1] != 'He'):
            raise NotImplementedError('[TODO] chromo_fluid_list for other elements list?')
        H = elements.get('H')
        He = elements.get('He')
        other = elements[2:]
        result = [SpecieList.value_type.electron(),
                  H.neutral(),
                  H.ion(1),
                  He.ion(1),
                  ]
        if self.params.get('do_helium', False):
            result.append(He.ion(2))  # include He_III
        result.extend(other.ion_list(q=1))
        return SpecieList(result, istart=0)  # istart=0 --> renumber for this list.
